/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 * vi:set noexpandtab tabstop=8 shiftwidth=8:
 *
 * Copyright (C) 2024 Red Hat <www.redhat.com>
 *
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <config.h>

#include <gnome-software.h>

#include "gs-dnf5-generated.h"

#include "gs-dnf5-progress-helper.h"

struct _GsDnf5ProgressHelper
{
	GObject		 parent;

	GsDnf5RpmRpm	*rpm_proxy;  /* (owned) */
	GsDnf5Base	*base_proxy;  /* (owned) */
	GsApp		*app; /* (owned) (nullable) */
	GsAppList	*list; /* (owned) (nullable) */
	gchar		*session_object_path; /* (owned) */
	GHashTable	*ongoing_downloads; /* (owned); gchar * (download_id) ~> DownloadInfo * */
	GsAppState	 recover_download_state;
	guint64		 to_download_total;
	guint64		 downloaded_total;

	gulong		 download_add_new_id;
	gulong		 download_progress_id;
	gulong		 download_mirror_failure_id;
	gulong		 download_end_id;
	gulong		 transaction_before_begin_id;
	gulong		 transaction_after_complete_id;
	gulong		 transaction_elem_progress_id;
	gulong		 transaction_action_progress_id;
	gulong		 transaction_action_start_id;
	gulong		 transaction_action_stop_id;
	gulong		 transaction_script_error_id;
	gulong		 transaction_script_start_id;
	gulong		 transaction_script_stop_id;
	gulong		 transaction_transaction_progress_id;
	gulong		 transaction_transaction_start_id;
	gulong		 transaction_transaction_stop_id;
	gulong		 transaction_unpack_error_id;
	gulong		 transaction_verify_progress_id;
	gulong		 transaction_verify_start_id;
	gulong		 transaction_verify_stop_id;
};

G_DEFINE_TYPE (GsDnf5ProgressHelper, gs_dnf5_progress_helper, G_TYPE_OBJECT)

static void
gs_dnf5_helper_set_progress (GsDnf5ProgressHelper *self,
			     guint progress)
{
	if (self->app != NULL)
		gs_app_set_progress (self->app, progress);
	if (self->list != NULL)
		gs_app_list_override_progress (self->list, progress);
}

static void
gs_dnf5_update_download_progress (GsDnf5ProgressHelper *self)
{
	if (self->to_download_total > 0)
		gs_dnf5_helper_set_progress (self, 100 * self->downloaded_total / self->to_download_total);
}

typedef struct {
	guint64 downloaded;
	guint64 total;
} DownloadInfo;

static void
gs_dnf5_download_add_new_cb (GsDnf5Base *object,
			     const gchar *arg_session_object_path,
			     const gchar *arg_download_id,
			     const gchar *arg_description,
			     gint64 arg_total_to_download,
			     GsDnf5ProgressHelper *self)
{
	DownloadInfo *download_info;

	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: download_id:'%s' description:'%s' total_to_download:%" G_GINT64_FORMAT,
		 G_STRFUNC, arg_download_id, arg_description, arg_total_to_download);
	g_return_if_fail (arg_download_id != NULL);

	if (arg_total_to_download <= 0)
		return;

	if (g_hash_table_size (self->ongoing_downloads) == 0) {
		if (self->app != NULL) {
			self->recover_download_state = gs_app_get_state (self->app);
			gs_app_set_state (self->app, GS_APP_STATE_DOWNLOADING);
		}
		self->to_download_total = 0;
		self->downloaded_total = 0;
	}

	download_info = g_new0 (DownloadInfo, 1);
	download_info->total = arg_total_to_download;

	g_hash_table_insert (self->ongoing_downloads, g_strdup (arg_download_id), download_info);
	self->to_download_total += arg_total_to_download;

	if (self->app != NULL)
		gs_app_set_size_download (self->app, GS_SIZE_TYPE_VALID, self->to_download_total);
	gs_dnf5_update_download_progress (self);
}

static void
gs_dnf5_download_end_cb (GsDnf5Base *object,
			 const gchar *arg_session_object_path,
			 const gchar *arg_download_id,
			 guint arg_transfer_status,
			 const gchar *arg_message,
			 GsDnf5ProgressHelper *self)
{
	DownloadInfo *download_info;

	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: download_id:'%s' transfer_status:%u message:'%s'",
		 G_STRFUNC, arg_download_id, arg_transfer_status, arg_message);
	g_return_if_fail (arg_download_id != NULL);

	download_info = g_hash_table_lookup (self->ongoing_downloads, arg_download_id);
	if (download_info != NULL) {
		self->downloaded_total -= download_info->downloaded;
		self->downloaded_total += download_info->total;

		/* remove after being done with the `download_info`, because it frees it */
		g_hash_table_remove (self->ongoing_downloads, arg_download_id);
		download_info = NULL;

		/* it's the last package to be downloaded */
		if (g_hash_table_size (self->ongoing_downloads) == 0) {
			gs_dnf5_helper_set_progress (self, GS_APP_PROGRESS_UNKNOWN);
			if (self->app != NULL) {
				gs_app_set_state (self->app, self->recover_download_state);
				gs_app_set_size_download (self->app, GS_SIZE_TYPE_UNKNOWN, 0);
			}
		} else {
			gs_dnf5_update_download_progress (self);
		}
	}
}

static void
gs_dnf5_download_mirror_failure_cb (GsDnf5Base *object,
				    const gchar *arg_session_object_path,
				    const gchar *arg_download_id,
				    const gchar *arg_message,
				    const gchar *arg_url,
				    const gchar *arg_metadata,
				    GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: download_id:'%s' message:'%s' url:'%s' metadata:'%s'",
		 G_STRFUNC, arg_download_id, arg_message, arg_url, arg_metadata);
}

static void
gs_dnf5_download_progress_cb (GsDnf5Base *object,
			      const gchar *arg_session_object_path,
			      const gchar *arg_download_id,
			      gint64 arg_total_to_download,
			      gint64 arg_downloaded,
			      GsDnf5ProgressHelper *self)
{
	DownloadInfo *download_info;

	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: download_id:'%s' total_to_download:%" G_GINT64_FORMAT " downloaded:%" G_GINT64_FORMAT,
		 G_STRFUNC, arg_download_id, arg_total_to_download, arg_downloaded);

	download_info = g_hash_table_lookup (self->ongoing_downloads, arg_download_id);
	if (download_info != NULL) {
		self->downloaded_total += arg_downloaded - download_info->downloaded;
		download_info->downloaded = arg_downloaded;
		if (download_info->downloaded != (guint64) arg_total_to_download) {
			self->to_download_total -= download_info->total;
			download_info->total = arg_total_to_download;
			self->to_download_total += download_info->total;
			if (self->app != NULL)
				gs_app_set_size_download (self->app, GS_SIZE_TYPE_VALID, self->to_download_total);
		}
		gs_dnf5_update_download_progress (self);
	}
}

static void
gs_dnf5_transaction_before_begin_cb (GsDnf5RpmRpm *object,
				     const gchar *arg_session_object_path,
				     guint64 arg_total,
				     GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	gs_dnf5_helper_set_progress (self, 0);
	g_debug ("%s: total:%" G_GUINT64_FORMAT, G_STRFUNC, arg_total);
}

static void
gs_dnf5_transaction_after_complete_cb (GsDnf5RpmRpm *object,
				       const gchar *arg_session_object_path,
				       gboolean arg_success,
				       GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: success:%d", G_STRFUNC, arg_success);
	gs_dnf5_helper_set_progress (self, 100);
}

static void
gs_dnf5_transaction_elem_progress_cb (GsDnf5RpmRpm *object,
				      const gchar *arg_session_object_path,
				      const gchar *arg_nevra,
				      guint64 arg_processed,
				      guint64 arg_total,
				      GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: nevra:'%s' progress:%" G_GUINT64_FORMAT "/%" G_GUINT64_FORMAT, G_STRFUNC, arg_nevra, arg_processed, arg_total);
	if (arg_total != 0)
		gs_dnf5_helper_set_progress (self, arg_processed * 100 / arg_total);
}

static void
gs_dnf5_transaction_action_progress_cb (GsDnf5RpmRpm *object,
					const gchar *arg_session_object_path,
					const gchar *arg_nevra,
					guint64 arg_processed,
					guint64 arg_total,
					GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: nevra:'%s' progress:%" G_GUINT64_FORMAT "/%" G_GUINT64_FORMAT, G_STRFUNC, arg_nevra, arg_processed, arg_total);
}

static void
gs_dnf5_transaction_action_start_cb (GsDnf5RpmRpm *object,
				     const gchar *arg_session_object_path,
				     const gchar *arg_nevra,
				     guint arg_action,
				     guint64 arg_total,
				     GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: nevra:'%s' action:%u total:%" G_GUINT64_FORMAT, G_STRFUNC, arg_nevra, arg_action, arg_total);
}

static void
gs_dnf5_transaction_action_stop_cb (GsDnf5RpmRpm *object,
				    const gchar *arg_session_object_path,
				    const gchar *arg_nevra,
				    guint64 arg_total,
				    GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: nevra:'%s' total:%" G_GUINT64_FORMAT, G_STRFUNC, arg_nevra, arg_total);
}

static void
gs_dnf5_transaction_script_error_cb (GsDnf5RpmRpm *object,
				     const gchar *arg_session_object_path,
				     const gchar *arg_nevra,
				     guint arg_scriptlet_type,
				     guint64 arg_return_code,
				     GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: nevra:'%s' scriptlet-type:%u return-code:%" G_GUINT64_FORMAT, G_STRFUNC, arg_nevra, arg_scriptlet_type, arg_return_code);
}

static void
gs_dnf5_transaction_script_start_cb (GsDnf5RpmRpm *object,
				     const gchar *arg_session_object_path,
				     const gchar *arg_nevra,
				     guint arg_scriptlet_type,
				     GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: nevra:'%s' scriptlet-type:%u", G_STRFUNC, arg_nevra, arg_scriptlet_type);
}

static void
gs_dnf5_transaction_script_stop_cb (GsDnf5RpmRpm *object,
				    const gchar *arg_session_object_path,
				    const gchar *arg_nevra,
				    guint arg_scriptlet_type,
				    guint64 arg_return_code,
				    GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: nevra:'%s' scriptlet-type:%u return-code:%" G_GUINT64_FORMAT, G_STRFUNC, arg_nevra, arg_scriptlet_type, arg_return_code);
}

static void
gs_dnf5_transaction_transaction_progress_cb (GsDnf5RpmRpm *object,
					     const gchar *arg_session_object_path,
					     guint64 arg_processed,
					     guint64 arg_total,
					     GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: progress:%" G_GUINT64_FORMAT "/%" G_GUINT64_FORMAT, G_STRFUNC, arg_processed, arg_total);
}

static void
gs_dnf5_transaction_transaction_start_cb (GsDnf5RpmRpm *object,
					  const gchar *arg_session_object_path,
					  guint64 arg_total,
				          GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: total:%" G_GUINT64_FORMAT, G_STRFUNC, arg_total);
}

static void
gs_dnf5_transaction_transaction_stop_cb (GsDnf5RpmRpm *object,
					 const gchar *arg_session_object_path,
					 guint64 arg_total,
				         GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: total:%" G_GUINT64_FORMAT, G_STRFUNC, arg_total);
}

static void
gs_dnf5_transaction_unpack_error_cb (GsDnf5RpmRpm *object,
				     const gchar *arg_session_object_path,
				     const gchar *arg_nevra,
				     GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: nevra:'%s'", G_STRFUNC, arg_nevra);
}

static void
gs_dnf5_transaction_verify_progress_cb (GsDnf5RpmRpm *object,
					const gchar *arg_session_object_path,
					guint64 arg_processed,
					guint64 arg_total,
					GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: progress:%" G_GUINT64_FORMAT "/%" G_GUINT64_FORMAT, G_STRFUNC, arg_processed, arg_total);
}

static void
gs_dnf5_transaction_verify_start_cb (GsDnf5RpmRpm *object,
				     const gchar *arg_session_object_path,
				     guint64 arg_total,
				     GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: total:%" G_GUINT64_FORMAT, G_STRFUNC, arg_total);
}

static void
gs_dnf5_transaction_verify_stop_cb (GsDnf5RpmRpm *object,
				    const gchar *arg_session_object_path,
				    guint64 arg_total,
				    GsDnf5ProgressHelper *self)
{
	if (g_strcmp0 (self->session_object_path, arg_session_object_path) != 0)
		return;
	g_debug ("%s: total:%" G_GUINT64_FORMAT, G_STRFUNC, arg_total);
}

static void
gs_dnf5_progress_helper_dispose (GObject *object)
{
	GsDnf5ProgressHelper *self = GS_DNF5_PROGRESS_HELPER (object);
	GObject *base_proxy = G_OBJECT (self->base_proxy);
	GObject *rpm_proxy = G_OBJECT (self->rpm_proxy);

	#define clear_signal_handler(_proxy, _id) G_STMT_START { \
		if (_id) { \
			g_signal_handler_disconnect (_proxy, _id); \
			_id = 0; \
		} \
		} G_STMT_END

	clear_signal_handler (base_proxy, self->download_add_new_id);
	clear_signal_handler (base_proxy, self->download_progress_id);
	clear_signal_handler (base_proxy, self->download_mirror_failure_id);
	clear_signal_handler (base_proxy, self->download_end_id);

	clear_signal_handler (rpm_proxy, self->transaction_before_begin_id);
	clear_signal_handler (rpm_proxy, self->transaction_after_complete_id);
	clear_signal_handler (rpm_proxy, self->transaction_elem_progress_id);
	clear_signal_handler (rpm_proxy, self->transaction_action_progress_id);
	clear_signal_handler (rpm_proxy, self->transaction_action_start_id);
	clear_signal_handler (rpm_proxy, self->transaction_action_stop_id);
	clear_signal_handler (rpm_proxy, self->transaction_script_error_id);
	clear_signal_handler (rpm_proxy, self->transaction_script_start_id);
	clear_signal_handler (rpm_proxy, self->transaction_script_stop_id);
	clear_signal_handler (rpm_proxy, self->transaction_transaction_progress_id);
	clear_signal_handler (rpm_proxy, self->transaction_transaction_start_id);
	clear_signal_handler (rpm_proxy, self->transaction_transaction_stop_id);
	clear_signal_handler (rpm_proxy, self->transaction_unpack_error_id);
	clear_signal_handler (rpm_proxy, self->transaction_verify_progress_id);
	clear_signal_handler (rpm_proxy, self->transaction_verify_start_id);
	clear_signal_handler (rpm_proxy, self->transaction_verify_stop_id);

	#undef clear_signal_handler

	gs_dnf5_helper_set_progress (self, GS_APP_PROGRESS_UNKNOWN);

	G_OBJECT_CLASS (gs_dnf5_progress_helper_parent_class)->dispose (object);
}

static void
gs_dnf5_progress_helper_finalize (GObject *object)
{
	GsDnf5ProgressHelper *self = GS_DNF5_PROGRESS_HELPER (object);

	g_clear_object (&self->base_proxy);
	g_clear_object (&self->rpm_proxy);
	g_clear_object (&self->app);
	g_clear_object (&self->list);
	g_clear_pointer (&self->session_object_path, g_free);
	g_clear_pointer (&self->ongoing_downloads, g_hash_table_destroy);

	G_OBJECT_CLASS (gs_dnf5_progress_helper_parent_class)->finalize (object);
}

static void
gs_dnf5_progress_helper_class_init (GsDnf5ProgressHelperClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->dispose = gs_dnf5_progress_helper_dispose;
	object_class->finalize = gs_dnf5_progress_helper_finalize;
}

static void
gs_dnf5_progress_helper_init (GsDnf5ProgressHelper *self)
{
	self->ongoing_downloads = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
}

GsDnf5ProgressHelper *
gs_dnf5_progress_helper_new (GsDnf5Base *base_proxy,
			     GsDnf5RpmRpm *rpm_proxy,
			     GsApp *progress_app,
			     GsAppList *progress_list,
			     const gchar *session_object_path)
{
	GsDnf5ProgressHelper *self;

	g_return_val_if_fail (GS_DNF5_IS_BASE (base_proxy), NULL);
	g_return_val_if_fail (GS_DNF5_IS_RPM_RPM (rpm_proxy), NULL);
	g_return_val_if_fail (GS_IS_APP (progress_app) || GS_IS_APP_LIST (progress_list), NULL);
	g_return_val_if_fail (session_object_path != NULL, NULL);

	self = g_object_new (GS_TYPE_DNF5_PROGRESS_HELPER, NULL);
	self->base_proxy = g_object_ref (base_proxy);
	self->rpm_proxy = g_object_ref (rpm_proxy);
	self->app = progress_app ? g_object_ref (progress_app) : NULL;
	self->list = progress_list ? g_object_ref (progress_list) : NULL;
	self->session_object_path = g_strdup (session_object_path);

	gs_dnf5_helper_set_progress (self, GS_APP_PROGRESS_UNKNOWN);

	self->download_add_new_id =
		g_signal_connect_object (base_proxy, "download_add_new",
					 G_CALLBACK (gs_dnf5_download_add_new_cb), self, 0);
	self->download_progress_id =
		g_signal_connect_object (base_proxy, "download_progress",
					 G_CALLBACK (gs_dnf5_download_progress_cb), self, 0);
	self->download_mirror_failure_id =
		g_signal_connect_object (base_proxy, "download_mirror_failure",
					 G_CALLBACK (gs_dnf5_download_mirror_failure_cb), self, 0);
	self->download_end_id =
		g_signal_connect_object (base_proxy, "download_end",
					 G_CALLBACK (gs_dnf5_download_end_cb), self, 0);

	self->transaction_before_begin_id =
		g_signal_connect_object (rpm_proxy, "transaction_before_begin",
					 G_CALLBACK (gs_dnf5_transaction_before_begin_cb), self, 0);
	self->transaction_after_complete_id =
		g_signal_connect_object (rpm_proxy, "transaction_after_complete",
					 G_CALLBACK (gs_dnf5_transaction_after_complete_cb), self, 0);
	self->transaction_elem_progress_id =
		g_signal_connect_object (rpm_proxy, "transaction_elem_progress",
					 G_CALLBACK (gs_dnf5_transaction_elem_progress_cb), self, 0);
	self->transaction_action_progress_id =
		g_signal_connect_object (rpm_proxy, "transaction_action_progress",
					 G_CALLBACK (gs_dnf5_transaction_action_progress_cb), self, 0);
	self->transaction_action_start_id =
		g_signal_connect_object (rpm_proxy, "transaction_action_start",
					 G_CALLBACK (gs_dnf5_transaction_action_start_cb), self, 0);
	self->transaction_action_stop_id =
		g_signal_connect_object (rpm_proxy, "transaction_action_stop",
					 G_CALLBACK (gs_dnf5_transaction_action_stop_cb), self, 0);
	self->transaction_script_error_id =
		g_signal_connect_object (rpm_proxy, "transaction_script_error",
					 G_CALLBACK (gs_dnf5_transaction_script_error_cb), self, 0);
	self->transaction_script_start_id =
		g_signal_connect_object (rpm_proxy, "transaction_script_start",
					 G_CALLBACK (gs_dnf5_transaction_script_start_cb), self, 0);
	self->transaction_script_stop_id =
		g_signal_connect_object (rpm_proxy, "transaction_script_stop",
					 G_CALLBACK (gs_dnf5_transaction_script_stop_cb), self, 0);
	self->transaction_transaction_progress_id =
		g_signal_connect_object (rpm_proxy, "transaction_transaction_progress",
					 G_CALLBACK (gs_dnf5_transaction_transaction_progress_cb), self, 0);
	self->transaction_transaction_start_id =
		g_signal_connect_object (rpm_proxy, "transaction_transaction_start",
					 G_CALLBACK (gs_dnf5_transaction_transaction_start_cb), self, 0);
	self->transaction_transaction_stop_id =
		g_signal_connect_object (rpm_proxy, "transaction_transaction_stop",
					 G_CALLBACK (gs_dnf5_transaction_transaction_stop_cb), self, 0);
	self->transaction_unpack_error_id =
		g_signal_connect_object (rpm_proxy, "transaction_unpack_error",
					 G_CALLBACK (gs_dnf5_transaction_unpack_error_cb), self, 0);
	self->transaction_verify_progress_id =
		g_signal_connect_object (rpm_proxy, "transaction_verify_progress",
					 G_CALLBACK (gs_dnf5_transaction_verify_progress_cb), self, 0);
	self->transaction_verify_start_id =
		g_signal_connect_object (rpm_proxy, "transaction_verify_start",
					 G_CALLBACK (gs_dnf5_transaction_verify_start_cb), self, 0);
	self->transaction_verify_stop_id =
		g_signal_connect_object (rpm_proxy, "transaction_verify_stop",
					 G_CALLBACK (gs_dnf5_transaction_verify_stop_cb), self, 0);

	return self;
}
