/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 * vi:set noexpandtab tabstop=8 shiftwidth=8:
 *
 * Copyright (C) 2024 Red Hat <www.redhat.com>
 *
 * SPDX-License-Identifier: GPL-2.0+
 */

/*
 * SECTION: dnf5 plugin.
 *
 * Talks to dnf5 to manage packages through dnf.
 */

#include <config.h>

#include <glib/gi18n-lib.h>
#include <gnome-software.h>

#include "gs-app-private.h"

#include "gs-worker-thread.h"
#include "gs-dnf5-generated.h"
#include "gs-dnf5-rpm-generated.h"
#include "gs-dnf5-progress-helper.h"

#include "gs-plugin-dnf5.h"

#define GS_DNF5_RELEASEVER_DEFAULT		NULL
#define GS_DNF5_INTERFACE_RPM_DNF		"org.rpm.dnf.v0"
#define GS_DNF5_OBJECT_PATH_RPM_DNF		"/org/rpm/dnf/v0"

/* for how long a session can be left opened until it's auto-closed */
#define GS_SESSION_LIFETIME_SECS (5 * 60)

struct _GsPluginDnf5
{
	GsPlugin		 parent;

	GsWorkerThread		*worker;  /* (owned) */
	GDBusConnection		*connection;
	GsDnf5RpmTransaction	*rpm_transaction_proxy;
	guint			 rpm_transaction_watch_id;
	gint			 calling_rpm;

	struct _SessionData {
		GMutex			 mutex;
		GCond			 cond;
		guint			 autoclose_timer;
		GsDnf5SessionManager	*proxy;
		gchar			*object_path;
		guint			 n_used;
		gint			 needs_reset;
	} session_data;

	struct _DependencySizesData {
		GMutex			 mutex;
		GPtrArray		*apps; /* (element-type GsApp) */
		GCancellable		*cancellable; /* non-NULL, when an op is running */
	} dependency_sizes_data;
};

G_DEFINE_TYPE (GsPluginDnf5, gs_plugin_dnf5, GS_TYPE_PLUGIN)

#define assert_in_worker(self) \
	g_assert (gs_worker_thread_is_in_worker_context (self->worker))

static gint
gs_dnf5_get_priority_for_interactivity (gboolean interactive)
{
	return interactive ? G_PRIORITY_DEFAULT : G_PRIORITY_LOW;
}

static void
gs_dnf5_convert_error (GError **error)
{
	gboolean not_authorized;
	if (error == NULL || *error == NULL)
		return;

	/* The message is not localized in the dnf5, thus this should work */
	not_authorized = strstr ((*error)->message, "GDBus.Error:org.rpm.dnf.v0.Error: Not authorized") != NULL;

	g_dbus_error_strip_remote_error (*error);
	gs_utils_error_convert_gdbus (error);
	if (not_authorized && g_error_matches (*error, G_IO_ERROR, G_IO_ERROR_DBUS_ERROR)) {
		(*error)->domain = GS_PLUGIN_ERROR;
		(*error)->code = GS_PLUGIN_ERROR_AUTH_REQUIRED;
	} else if (g_error_matches (*error, G_IO_ERROR, G_IO_ERROR_DBUS_ERROR) || (
		   !gs_utils_error_convert_gdbus (error) &&
		   !gs_utils_error_convert_gio (error))) {
		(*error)->domain = GS_PLUGIN_ERROR;
		(*error)->code = GS_PLUGIN_ERROR_FAILED;
	}
}

static void
gs_dnf5_report_error (GsPluginDnf5 *self,
		      GsPluginEventCallback event_callback,
		      void *event_user_data,
		      GsApp *app,
		      const GError *error,
		      gboolean interactive)
{
	g_autoptr(GsPluginEvent) event = NULL;

	if (!event_callback)
		return;

	event = gs_plugin_event_new ("error", error,
				     app ? "app" : NULL, app,
				     NULL);
	if (interactive)
		gs_plugin_event_add_flag (event, GS_PLUGIN_EVENT_FLAG_INTERACTIVE);
	gs_plugin_event_add_flag (event, GS_PLUGIN_EVENT_FLAG_WARNING);
	event_callback (GS_PLUGIN (self), event, event_user_data);
}

/* The session reset is used to refresh information about the packages,
   like after a package install/uninstall, it might not be noticed by
   the daemon internal structures that the package is installed/uninstalled. */
static void
gs_dnf5_mark_session_needs_reset (GsPluginDnf5 *self)
{
	g_atomic_int_set (&self->session_data.needs_reset, 1);
}

static void
gs_dnf5_rpm_start_transaction_cb (GsDnf5RpmTransaction *proxy,
				  const gchar *dbcookie,
				  guint transaction_id,
				  gpointer user_data)
{
	GsPluginDnf5 *self = user_data;

	g_debug ("%s: %p: dbcookie:'%s' transaction_id:%u", G_STRFUNC, self, dbcookie, transaction_id);
}

static void
gs_dnf5_rpm_end_transaction_cb (GsDnf5RpmTransaction *proxy,
				const gchar *dbcookie,
				guint transaction_id,
				gpointer user_data)
{
	GsPluginDnf5 *self = user_data;

	g_debug ("%s: %p: dbcookie:'%s' transaction_id:%u", G_STRFUNC, self, dbcookie, transaction_id);

	/* the transactions are for install/uninstall, thus
	   mark the connection to require a reset() */
	gs_dnf5_mark_session_needs_reset (self);

	/* also do not know what precisely changed, thus a heavy hammer */
	gs_plugin_reload (GS_PLUGIN (self));
}

static void
gs_dnf5_got_rpm_transaction_proxy_cb (GObject *source_object,
				      GAsyncResult *result,
				      gpointer user_data)
{
	g_autoptr(GsPluginDnf5) self = user_data;
	g_autoptr(GError) local_error = NULL;

	self->rpm_transaction_proxy = gs_dnf5_rpm_transaction_proxy_new_finish (result, &local_error);

	if (self->rpm_transaction_proxy != NULL) {
		g_debug ("Opened RPM Transaction D-Bus proxy");

		g_signal_connect_object (self->rpm_transaction_proxy, "start-transaction",
					 G_CALLBACK (gs_dnf5_rpm_start_transaction_cb), self, 0);
		g_signal_connect_object (self->rpm_transaction_proxy, "end-transaction",
					 G_CALLBACK (gs_dnf5_rpm_end_transaction_cb), self, 0);
	} else {
		g_debug ("Failed to open RPM Transaction proxy: %s", local_error ? local_error->message : "Unknown error");
	}
}

static void
gs_dnf5_rpm_transaction_appeared_cb (GDBusConnection *connection,
				     const gchar *name,
				     const gchar *name_owner,
				     gpointer user_data)
{
	GsPluginDnf5 *self = user_data;

	g_debug ("%s: name:'%s' owner:'%s' existing-proxy:%p calling_rpm:%d",
		 G_STRFUNC, name, name_owner, self->rpm_transaction_proxy,
		 g_atomic_int_get (&self->calling_rpm));

	if (name_owner != NULL && *name_owner != '\0' && !g_atomic_int_get (&self->calling_rpm)) {
		g_clear_object (&self->rpm_transaction_proxy);
		gs_dnf5_rpm_transaction_proxy_new (self->connection,
						   G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						   "org.rpm.announce",
						   "/org/rpm/Transaction",
						   NULL,
						   gs_dnf5_got_rpm_transaction_proxy_cb,
						   g_object_ref (self));
	}
}

static void
gs_dnf5_rpm_transaction_vanished_cb (GDBusConnection *connection,
				     const gchar *name,
				     gpointer user_data)
{
	GsPluginDnf5 *self = user_data;

	g_debug ("%s: name:'%s' existing-proxy:%p", G_STRFUNC, name, self->rpm_transaction_proxy);

	g_clear_object (&self->rpm_transaction_proxy);
}

static void
gs_dnf5_bus_get_cb (GObject *source_object,
		    GAsyncResult *result,
		    gpointer user_data)
{
	g_autoptr(GTask) task = user_data;
	g_autoptr(GError) local_error = NULL;
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (g_task_get_source_object (task));

	self->connection = g_bus_get_finish (result, &local_error);
	gs_dnf5_convert_error (&local_error);

	if (self->connection != NULL) {
		self->rpm_transaction_watch_id = g_bus_watch_name_on_connection (self->connection,
										 "org.rpm.announce",
										 G_BUS_NAME_WATCHER_FLAGS_NONE,
										 gs_dnf5_rpm_transaction_appeared_cb,
										 gs_dnf5_rpm_transaction_vanished_cb,
										 self,
										 NULL);
	}

	if (self->connection != NULL)
		g_task_return_boolean (task, TRUE);
	else
		g_task_return_error (task, g_steal_pointer (&local_error));
}

static void
gs_plugin_dnf5_setup_async (GsPlugin *plugin,
			    GCancellable *cancellable,
			    GAsyncReadyCallback callback,
			    gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	g_autoptr(GTask) task = NULL;

	g_debug ("dnf5 setup");

	/* Start up a worker thread to process all the plugin’s function calls. */
	self->worker = gs_worker_thread_new ("gs-plugin-dnf5");

	task = g_task_new (plugin, cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_setup_async);

	g_bus_get (G_BUS_TYPE_SYSTEM, cancellable, gs_dnf5_bus_get_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_setup_finish (GsPlugin *plugin,
			     GAsyncResult *result,
			     GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

static void
gs_dnf5_shutdown_cb (GObject *source_object,
		     GAsyncResult *result,
		     gpointer user_data)
{
	g_autoptr(GTask) task = user_data;
	GsPluginDnf5 *self = g_task_get_source_object (task);
	g_autoptr(GsWorkerThread) worker = NULL;
	g_autoptr(GError) local_error = NULL;

	worker = g_steal_pointer (&self->worker);

	if (!gs_worker_thread_shutdown_finish (worker, result, &local_error))
		g_task_return_error (task, g_steal_pointer (&local_error));
	else
		g_task_return_boolean (task, TRUE);
}

static void
gs_plugin_dnf5_shutdown_async (GsPlugin *plugin,
			       GCancellable *cancellable,
			       GAsyncReadyCallback callback,
			       gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	g_autoptr(GTask) task = NULL;
	g_autoptr(GCancellable) dependency_sizes_cancellable = NULL;

	task = g_task_new (plugin, cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_shutdown_async);

	/* stop any dependency sizes ongoing operation */
	g_mutex_lock (&self->dependency_sizes_data.mutex);
	if (self->dependency_sizes_data.cancellable)
		dependency_sizes_cancellable = g_object_ref (self->dependency_sizes_data.cancellable);
	g_mutex_unlock (&self->dependency_sizes_data.mutex);
	if (dependency_sizes_cancellable != NULL)
		g_cancellable_cancel (dependency_sizes_cancellable);

	/* Stop the worker thread. */
	gs_worker_thread_shutdown_async (self->worker, cancellable, gs_dnf5_shutdown_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_shutdown_finish (GsPlugin *plugin,
				GAsyncResult *result,
				GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

static void
gs_dnf5_close_session_real (GsDnf5SessionManager *proxy,
			    const gchar *session_path)
{
	g_autoptr(GError) local_error = NULL;
	gboolean result = FALSE;

	/* Not passing operation's cancellable, because it can be cancelled by the user,
	   but the session needs to be closed. */
	if (gs_dnf5_session_manager_call_close_session_sync (proxy, session_path, &result, NULL, &local_error))
		g_warn_if_fail (result);
	else
		g_debug ("Failed to close session: %s", local_error->message);
}

static gboolean
gs_dnf5_autoclose_session_cb (gpointer user_data)
{
	GsPluginDnf5 *self = user_data;
	g_autoptr(GMutexLocker) locker = g_mutex_locker_new (&self->session_data.mutex);

	self->session_data.autoclose_timer = 0;
	if (self->session_data.n_used == 0 && self->session_data.proxy != NULL) {
		gs_dnf5_close_session_real (self->session_data.proxy, self->session_data.object_path);

		g_clear_object (&self->session_data.proxy);
		g_clear_pointer (&self->session_data.object_path, g_free);
	}

	return G_SOURCE_REMOVE;
}

static gboolean
gs_dnf5_reset_session_sync (GsPluginDnf5 *self,
			    const gchar *session_path,
			    GCancellable *cancellable,
			    GError **error)
{
	g_autoptr(GsDnf5Base) base_proxy = NULL;
	g_autofree gchar *error_msg = NULL;
	gboolean success = FALSE;

	base_proxy = gs_dnf5_base_proxy_new_sync (self->connection,
						  G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						  GS_DNF5_INTERFACE_RPM_DNF,
						  session_path,
						  cancellable,
						  error);
	if (base_proxy == NULL) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to create Base proxy: ");
		return FALSE;
	}

	if (!gs_dnf5_base_call_reset_sync (base_proxy, &success, &error_msg, cancellable, error)) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to call Base::reset(): ");
		return FALSE;
	}

	if (!success) {
		g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "Failed to execute Base::reset(): %s", error_msg ? error_msg : "Unknown error");
		return FALSE;
	}

	return TRUE;
}

static gchar * /* (transfer full) */
gs_dnf5_open_session (GsPluginDnf5 *self,
		      const gchar *releasever,
		      GsDnf5SessionManager **out_proxy, /* (transfer full) */
		      GCancellable *cancellable,
		      GError **error)
{
	g_autoptr(GMutexLocker) locker = g_mutex_locker_new (&self->session_data.mutex);
	g_autoptr(GVariantBuilder) options_builder = NULL;
	g_autofree gchar *session_path = NULL;

	if (releasever == GS_DNF5_RELEASEVER_DEFAULT) {
		if (self->session_data.autoclose_timer) {
			g_source_remove (self->session_data.autoclose_timer);
			self->session_data.autoclose_timer = 0;
		}

		if (self->session_data.proxy != NULL) {
			GsDnf5SessionManager *proxy = self->session_data.proxy;

			while (g_atomic_int_get (&self->session_data.needs_reset) &&
			       !g_cancellable_is_cancelled (cancellable)) {
				g_autoptr(GError) local_error = NULL;
				if (self->session_data.n_used > 0) {
					g_debug ("Existing session needs reset, waiting for opened operations to finish");
					while (self->session_data.n_used && g_atomic_int_get (&self->session_data.needs_reset)) {
						g_cond_wait_until (&self->session_data.cond, &self->session_data.mutex,
								   /* check whether was cancelled up to three times per second */
								   g_get_monotonic_time () + (G_TIME_SPAN_SECOND / 3));

						if (g_cancellable_set_error_if_cancelled (cancellable, error))
							return NULL;
					}
				}

				if (g_cancellable_set_error_if_cancelled (cancellable, error))
					return NULL;

				if (g_atomic_int_get (&self->session_data.needs_reset)) {
					gboolean success;

					success = gs_dnf5_reset_session_sync (self, self->session_data.object_path, cancellable, &local_error);
					g_atomic_int_set (&self->session_data.needs_reset, 0);

					/* let any waiter for session reset know the conditions changed */
					g_cond_broadcast (&self->session_data.cond);

					if (success) {
						g_debug ("Successfully reset existing session");
					} else {
						g_debug ("Failed to reset existing session: %s", local_error ? local_error->message : "Unknown error");
						break;
					}
				} else {
					g_debug ("Wanted to reset the session, but someone else did that already, thus skipping the reset");
					break;
				}
			}

			if (g_cancellable_set_error_if_cancelled (cancellable, error))
				return NULL;

			if (proxy == self->session_data.proxy) {
				self->session_data.n_used++;
				g_debug ("Using existing session");
				*out_proxy = g_object_ref (self->session_data.proxy);
				return g_strdup (self->session_data.object_path);
			}
		}
	}

	g_debug ("Creating new session");
	*out_proxy = gs_dnf5_session_manager_proxy_new_sync (self->connection,
							     G_DBUS_PROXY_FLAGS_NONE,
							     GS_DNF5_INTERFACE_RPM_DNF,
							     GS_DNF5_OBJECT_PATH_RPM_DNF,
							     cancellable,
							     error);

	if (*out_proxy == NULL) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to get Session Manager: ");
		return NULL;
	}

	options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));

	if (releasever != GS_DNF5_RELEASEVER_DEFAULT) {
		g_variant_builder_add (options_builder, "{sv}", "releasever",
				       g_variant_new_string (releasever));
	} else {
		g_autoptr(GVariantBuilder) config = NULL;

		config = g_variant_builder_new (G_VARIANT_TYPE ("a{ss}"));
		g_variant_builder_add (config, "{ss}", "optional_metadata_types",
				       "appstream");

		g_variant_builder_add (options_builder, "{sv}", "config",
				       g_variant_builder_end (config));
	}

	if (!gs_dnf5_session_manager_call_open_session_sync (*out_proxy,
							     g_variant_builder_end (options_builder),
							     &session_path,
							     cancellable,
							     error)) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to open session: ");
		return NULL;
	}

	if (releasever == GS_DNF5_RELEASEVER_DEFAULT) {
		g_assert (self->session_data.proxy == NULL);
		self->session_data.proxy = g_object_ref (*out_proxy);
		self->session_data.object_path = g_strdup (session_path);
		self->session_data.n_used = 1;
	}

	return g_steal_pointer (&session_path);
}

static void
gs_dnf5_close_session (GsPluginDnf5 *self,
		       GsDnf5SessionManager *proxy,
		       const gchar *session_path)
{
	g_autoptr(GMutexLocker) locker = g_mutex_locker_new (&self->session_data.mutex);

	if (self->session_data.proxy == proxy) {
		g_assert (self->session_data.n_used > 0);
		self->session_data.n_used--;

		/* let any waiter for session reset know the n_used changed */
		g_cond_broadcast (&self->session_data.cond);

		if (self->session_data.n_used == 0) {
			g_assert (self->session_data.autoclose_timer == 0);
			self->session_data.autoclose_timer = g_timeout_add_seconds (GS_SESSION_LIFETIME_SECS, gs_dnf5_autoclose_session_cb, self);
		}
	} else {
		gs_dnf5_close_session_real (proxy, session_path);
	}
}

/* Returns whether to continue */
typedef gboolean (* GsDnf5ForeachFunc) (GsPluginDnf5 *self,
					GVariant *value,
					gpointer user_data,
					GCancellable *cancellable,
					GError **error);

static gboolean
gs_dnf5_foreach_item (GsPluginDnf5 *self,
		      GVariant *array,
		      GsDnf5ForeachFunc func,
		      gpointer user_data,
		      GCancellable *cancellable,
		      GError **error)
{
	GVariantIter iter;
	GVariant *child;

	g_variant_iter_init (&iter, array);
	while (child = g_variant_iter_next_value (&iter), child != NULL) {
		g_autoptr(GError) local_error = NULL;
		gboolean done;

		done = !func (self, child, user_data, cancellable, &local_error);
		g_variant_unref (child);

		if (local_error) {
			g_propagate_error (error, g_steal_pointer (&local_error));
			return FALSE;
		}

		if (g_cancellable_set_error_if_cancelled (cancellable, error))
			return FALSE;

		if (done)
			break;
	}

	return TRUE;
}

static void
gs_dnf5_app_set_str (GVariantDict *dict,
		     const gchar *key,
		     GsApp *app,
		     void (*set_func) (GsApp *app,
				       const gchar *value))
{
	g_autoptr(GVariant) value = NULL;

	value = g_variant_dict_lookup_value (dict, key, G_VARIANT_TYPE_STRING);
	if (value != NULL)
		set_func (app, g_variant_get_string (value, NULL));
}

static void
gs_dnf5_app_set_str2 (GVariantDict *dict,
		      const gchar *key,
		      GsApp *app,
		      void (*set_func) (GsApp *app,
					GsAppQuality quality,
					const gchar *value))
{
	g_autoptr(GVariant) value = NULL;

	value = g_variant_dict_lookup_value (dict, key, G_VARIANT_TYPE_STRING);
	if (value != NULL)
		set_func (app, GS_APP_QUALITY_NORMAL, g_variant_get_string (value, NULL));
}

static void
gs_dnf5_app_set_size (GVariantDict *dict,
		      const gchar *key,
		      GsApp *app,
		      void (*set_func) (GsApp *app,
					GsSizeType size_type,
					guint64 size_bytes))
{
	g_autoptr(GVariant) value = NULL;

	value = g_variant_dict_lookup_value (dict, key, G_VARIANT_TYPE_UINT64);
	if (value != NULL)
		set_func (app, GS_SIZE_TYPE_VALID, g_variant_get_uint64 (value));
}

static gchar *
gs_dnf5_dup_version_from_dict (GVariantDict *dict)
{
	g_autoptr(GVariant) version = NULL;

	version = g_variant_dict_lookup_value (dict, "version", G_VARIANT_TYPE_STRING);
	if (version != NULL) {
		const gchar *version_str = g_variant_get_string (version, NULL);
		if (version_str != NULL && *version_str != '\0') {
			g_autoptr(GVariant) release = NULL;
			const gchar *release_str = NULL;
			release = g_variant_dict_lookup_value (dict, "release", G_VARIANT_TYPE_STRING);
			if (release != NULL)
				release_str = g_variant_get_string (release, NULL);
			if (release_str != NULL && *release_str != '\0') {
				g_autofree gchar *version_release_str = g_strconcat (version_str, "-", release_str, NULL);
				return g_steal_pointer (&version_release_str);
			} else {
				return g_strdup (version_str);
			}
		}
	}

	return NULL;
}

static void
gs_dnf5_app_set_version (GVariantDict *dict,
			 GsApp *app,
			 void (*set_func) (GsApp *app,
					   const gchar *value))
{
	g_autofree gchar *ver_str = gs_dnf5_dup_version_from_dict (dict);
	if (ver_str != NULL)
		set_func (app, ver_str);
}

static void
gs_plugin_dnf5_set_packaging_format (GsApp *app)
{
	gs_app_set_metadata (app, "GnomeSoftware::PackagingFormat", "RPM");
	gs_app_set_metadata (app, "GnomeSoftware::PackagingBaseCssColor", "error_color");
}

static void
gs_dnf5_update_app_state (GsApp *app,
			  GsAppState state,
			  GVariantDict *dict)
{
	if (state == GS_APP_STATE_UNKNOWN) {
		g_autoptr(GVariant) value = NULL;
		value = g_variant_dict_lookup_value (dict, "is_installed", G_VARIANT_TYPE_BOOLEAN);
		if (value == NULL)
			return;
		if (g_variant_get_boolean (value))
			state = GS_APP_STATE_INSTALLED;
		else
			state = GS_APP_STATE_AVAILABLE;
	}
	/* The state can be only installed, available or updatable  */
	if (gs_app_get_state (app) == state ||
	    gs_app_get_state (app) == GS_APP_STATE_UPDATABLE)
		return;
	if (gs_app_get_state (app) == GS_APP_STATE_UNKNOWN)
		gs_app_set_state (app, state);
	else if (gs_app_get_state (app) != GS_APP_STATE_INSTALLED ||
		 state == GS_APP_STATE_UPDATABLE)
		gs_app_set_state (app, state);
}

static void
gs_dnf5_update_app_changelogs (GsApp *app,
			       GVariantDict *dict)
{
	GVariantIter iter;
	GVariant *child;
	g_autoptr(GString) changes = NULL;
	g_autoptr(GVariant) changelogs = NULL;

	changelogs = g_variant_dict_lookup_value (dict, "changelogs", NULL);
	if (changelogs == NULL)
		return;

	g_variant_iter_init (&iter, changelogs);
	while (child = g_variant_iter_next_value (&iter), child != NULL) {
		gint64 when = 0;
		const gchar *whom = NULL;
		const gchar *what = NULL;

		g_variant_get (child, "(x&s&s)", &when, &whom, &what);

		if (whom != NULL && what != NULL) {
			g_autofree gchar *tmp = NULL;
			const gchar *eml_start, *eml_end;

			/* Hide the email address in the changes */
			eml_start = strchr (whom, '<');
			eml_end = strrchr (whom, '>');
			if (eml_start != NULL && eml_start < eml_end) {
				if (g_ascii_isspace (eml_end[1]))
					eml_end++;
				tmp = g_malloc0 (strlen (whom) + 1);
				strncpy (tmp, whom, eml_start - whom);
				strcat (tmp, eml_end + 1);
				whom = tmp;
			}

			if (changes == NULL)
				changes = g_string_new ("");
			else
				g_string_append (changes, "\n\n");
			g_string_append (changes, whom);
			g_string_append_c (changes, '\n');
			g_string_append (changes, what);
		}

		g_variant_unref (child);
	}

	if (changes != NULL)
		gs_app_set_update_details_text (app, changes->str);
}

typedef struct {
	GsAppList *list;
	GsAppState set_state;
	GHashTable *nevra_to_app; /* (nullable): gchar *nevra ~> GsApp * */
	GHashTable *replaces; /* (nullable): GINT_TO_POINTER(pkg-id) ~> gchar *old_version */
	GHashTable *used_replaces; /* (nullable): GINT_TO_POINTER(pkg-id) ~> NULL */
	gboolean cover_unused_replaces;
} ReadPackageData;

static gboolean
gs_dnf5_read_package_cb (GsPluginDnf5 *self,
			 GVariant *package_array,
			 gpointer user_data,
			 GCancellable *cancellable,
			 GError **error)
{
	ReadPackageData *rpd = user_data;
	GsAppList *list = rpd->list;
	GHashTable *nevra_to_app = rpd->nevra_to_app;
	g_autofree gchar *name = NULL;
	g_autoptr(GsApp) app = NULL;
	g_autoptr(GVariantDict) dict = NULL;
	g_autoptr(GVariant) value = NULL;

	dict = g_variant_dict_new (package_array);

	value = g_variant_dict_lookup_value (dict, "name", G_VARIANT_TYPE_STRING);
	if (value != NULL) {
		const gchar *name_const = g_variant_get_string (value, NULL);
		app = gs_plugin_cache_lookup (GS_PLUGIN (self), name_const);
		if (app == NULL)
			name = g_strdup (name_const);
		g_clear_pointer (&value, g_variant_unref);
	}

	if (app == NULL) {
		app = gs_app_new (NULL);
		gs_app_set_management_plugin (app, GS_PLUGIN (self));
		gs_app_set_metadata (app, "GnomeSoftware::Creator", gs_plugin_get_name (GS_PLUGIN (self)));
		gs_plugin_dnf5_set_packaging_format (app);
		gs_app_set_kind (app, AS_COMPONENT_KIND_GENERIC);
		gs_app_set_bundle_kind (app, AS_BUNDLE_KIND_PACKAGE);
		gs_app_set_scope (app, AS_COMPONENT_SCOPE_SYSTEM);

		gs_dnf5_app_set_str (dict, "name", app, gs_app_add_source);
		gs_dnf5_app_set_str (dict, "nevra", app, gs_app_add_source_id);
		gs_dnf5_app_set_str2 (dict, "name", app, gs_app_set_name);
		gs_dnf5_app_set_str2 (dict, "summary", app, gs_app_set_summary);
		gs_dnf5_app_set_str2 (dict, "description", app, gs_app_set_description);
		gs_dnf5_app_set_str2 (dict, "license", app, gs_app_set_license);
		gs_dnf5_app_set_size (dict, "install_size", app, gs_app_set_size_installed);
		gs_dnf5_app_set_size (dict, "download_size", app, gs_app_set_size_download);

		value = g_variant_dict_lookup_value (dict, "url", G_VARIANT_TYPE_STRING);
		if (value != NULL) {
			gs_app_set_url (app, AS_URL_KIND_HOMEPAGE, g_variant_get_string (value, NULL));
			g_clear_pointer (&value, g_variant_unref);
		}

		gs_plugin_cache_add (GS_PLUGIN (self), name, app);
	}

	gs_dnf5_update_app_changelogs (app, dict);
	gs_dnf5_update_app_state (app, rpd->set_state, dict);

	if (rpd->set_state == GS_APP_STATE_UPDATABLE) {
		gs_dnf5_app_set_version (dict, app, gs_app_set_update_version);
		gs_app_add_quirk (app, GS_APP_QUIRK_NEEDS_REBOOT);
	} else {
		gs_dnf5_app_set_version (dict, app, gs_app_set_version);
	}

	if (nevra_to_app != NULL) {
		g_clear_pointer (&value, g_variant_unref);
		value = g_variant_dict_lookup_value (dict, "nevra", G_VARIANT_TYPE_STRING);
		if (value != NULL) {
			const gchar *nevra_const = g_variant_get_string (value, NULL);
			if (nevra_const != NULL && *nevra_const != '\0')
				g_hash_table_insert (nevra_to_app, g_strdup (nevra_const), g_object_ref (app));
		}
	}

	gs_app_list_add (list, app);

	return TRUE;
}

static gboolean
gs_dnf5_read_repo_cb (GsPluginDnf5 *self,
		      GVariant *package_array,
		      gpointer user_data,
		      GCancellable *cancellable,
		      GError **error)
{
	GsAppList *list = user_data;
	g_autoptr(GsApp) app = NULL;
	g_autoptr(GVariantDict) dict = NULL;
	g_autoptr(GVariant) value = NULL;
	const gchar *id;

	dict = g_variant_dict_new (package_array);

	value = g_variant_dict_lookup_value (dict, "id", G_VARIANT_TYPE_STRING);
	if (value == NULL)
		return TRUE;

	id = g_variant_get_string (value, NULL);
	if (id == NULL || *id == '\0')
		return TRUE;

	/* Skip sub-repos */
	if (g_str_has_suffix (id, "-source") ||
	    g_str_has_suffix (id, "-debuginfo") ||
	    g_str_has_suffix (id, "-testing"))
		return TRUE;

	g_clear_pointer (&value, g_variant_unref);
	id = NULL;

	app = gs_app_new (NULL);
	gs_app_set_management_plugin (app, GS_PLUGIN (self));
	gs_app_set_kind (app, AS_COMPONENT_KIND_REPOSITORY);
	gs_app_set_bundle_kind (app, AS_BUNDLE_KIND_PACKAGE);
	gs_app_set_scope (app, AS_COMPONENT_SCOPE_SYSTEM);
	gs_app_add_quirk (app, GS_APP_QUIRK_NOT_LAUNCHABLE);
	gs_app_set_metadata (app, "GnomeSoftware::Creator", gs_plugin_get_name (GS_PLUGIN (self)));
	gs_plugin_dnf5_set_packaging_format (app);
	gs_app_set_metadata (app, "GnomeSoftware::SortKey", "300");
	gs_app_set_origin_ui (app, _("Packages"));

	gs_dnf5_app_set_str (dict, "id", app, gs_app_set_id);
	gs_dnf5_app_set_str2 (dict, "name", app, gs_app_set_name);
	gs_dnf5_app_set_str2 (dict, "name", app, gs_app_set_description);

	value = g_variant_dict_lookup_value (dict, "enabled", G_VARIANT_TYPE_BOOLEAN);
	if (value != NULL) {
		gs_app_set_state (app, g_variant_get_boolean (value) ?
				  GS_APP_STATE_INSTALLED : GS_APP_STATE_AVAILABLE);
		g_clear_pointer (&value, g_variant_unref);
	} else {
		gs_app_set_state (app, GS_APP_STATE_INSTALLED);
	}

	gs_app_list_add (list, app);

	return TRUE;
}

static gboolean
gs_dnf5_refine_from_advisory_packages_cb (GsPluginDnf5 *self,
					  GVariant *array,
					  gpointer user_data,
					  GCancellable *cancellable,
					  GError **error)
{
	GHashTable *nevra_to_app = user_data;
	g_autoptr(GVariantDict) dict = NULL;
	g_autoptr(GVariant) var_nevra = NULL;

	dict = g_variant_dict_new (array);

	var_nevra = g_variant_dict_lookup_value (dict, "nevra", G_VARIANT_TYPE_STRING);
	if (var_nevra != NULL) {
		const gchar *nevra = g_variant_get_string (var_nevra, NULL);
		if (nevra != NULL) {
			GsApp *app;

			app = g_hash_table_lookup (nevra_to_app, nevra);
			if (app != NULL)
				gs_app_set_update_urgency (app, AS_URGENCY_KIND_CRITICAL);
		}
	}

	return TRUE;
}

static gboolean
gs_dnf5_refine_from_advisory_collections_cb (GsPluginDnf5 *self,
					     GVariant *array,
					     gpointer user_data,
					     GCancellable *cancellable,
					     GError **error)
{
	GHashTable *nevra_to_app = user_data;
	g_autoptr(GVariantDict) dict = NULL;
	g_autoptr(GVariant) packages = NULL;
	gboolean res = TRUE;

	dict = g_variant_dict_new (array);

	packages = g_variant_dict_lookup_value (dict, "packages", G_VARIANT_TYPE_ARRAY);
	if (packages != NULL)
		res = gs_dnf5_foreach_item (self, packages, gs_dnf5_refine_from_advisory_packages_cb, nevra_to_app, cancellable, error);

	return res;
}

static gboolean
gs_dnf5_refine_from_advisory_cb (GsPluginDnf5 *self,
				 GVariant *array,
				 gpointer user_data,
				 GCancellable *cancellable,
				 GError **error)
{
	GHashTable *nevra_to_app = user_data;
	g_autoptr(GVariantDict) dict = NULL;
	g_autoptr(GVariant) collections = NULL;
	g_autoptr(GVariant) severity = NULL;
	gboolean res = TRUE;

	dict = g_variant_dict_new (array);

	/* ensure only advisories with severity 'critical' are traversed */
	severity = g_variant_dict_lookup_value (dict, "severity", G_VARIANT_TYPE_STRING);
	if (severity == NULL || g_variant_get_string (severity, NULL) == NULL ||
	    g_ascii_strcasecmp (g_variant_get_string (severity, NULL), "critical") != 0)
		return TRUE;

	collections = g_variant_dict_lookup_value (dict, "collections", G_VARIANT_TYPE_ARRAY);
	if (collections != NULL)
		res = gs_dnf5_foreach_item (self, collections, gs_dnf5_refine_from_advisory_collections_cb, nevra_to_app, cancellable, error);

	return res;
}

static GVariant *
gs_dnf5_dup_package_attrs (void)
{
	g_autoptr(GVariantBuilder) attrs_builder = NULL;

	attrs_builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
	g_variant_builder_add (attrs_builder, "s", "nevra");
	g_variant_builder_add (attrs_builder, "s", "name");
	g_variant_builder_add (attrs_builder, "s", "epoch");
	g_variant_builder_add (attrs_builder, "s", "version");
	g_variant_builder_add (attrs_builder, "s", "release");
	g_variant_builder_add (attrs_builder, "s", "arch");
	g_variant_builder_add (attrs_builder, "s", "repo_id");
	g_variant_builder_add (attrs_builder, "s", "install_size");
	g_variant_builder_add (attrs_builder, "s", "download_size");
	g_variant_builder_add (attrs_builder, "s", "is_installed");
	g_variant_builder_add (attrs_builder, "s", "summary");
	g_variant_builder_add (attrs_builder, "s", "url");
	g_variant_builder_add (attrs_builder, "s", "license");
	g_variant_builder_add (attrs_builder, "s", "description");
	g_variant_builder_add (attrs_builder, "s", "changelogs");

	return g_variant_builder_end (attrs_builder);
}

static void
gs_dnf5_gather_async_result_cb (GObject *source_object,
				GAsyncResult *result,
				gpointer user_data)
{
	GAsyncResult **out_result = user_data;
	*out_result = g_object_ref (result);
}

static void
gs_dnf5_confirm_key_response_cb (GObject *source_object,
				 GAsyncResult *result,
				 gpointer user_data)
{
	g_autoptr(GError) local_error = NULL;

	if (!gs_dnf5_rpm_repo_call_confirm_key_finish (GS_DNF5_RPM_REPO (source_object), result, &local_error))
		g_debug ("Failed to confirm key: %s", local_error->message);
}

typedef struct {
	gchar *title;
	gchar *msg;
	gchar *details;
	gchar *accept_label;
	gchar *key_id;
	GsPluginDnf5 *self; /* (owned) */
	GWeakRef repo_proxy_weakref; /* GsDnf5RpmRepo */
} QuestionData;

static void
question_data_free (QuestionData *data)
{
	if (data != NULL) {
		g_free (data->title);
		g_free (data->msg);
		g_free (data->details);
		g_free (data->accept_label);
		g_free (data->key_id);
		g_clear_object (&data->self);
		g_weak_ref_clear (&data->repo_proxy_weakref);
		g_free (data);
	}
}

static gboolean
gs_dnf5_ask_question_idle_cb (gpointer user_data)
{
	QuestionData *data = user_data;
	g_autoptr(GsDnf5RpmRepo) repo_proxy = NULL;

	repo_proxy = g_weak_ref_get (&data->repo_proxy_weakref);
	if (repo_proxy != NULL) {
		gboolean confirmed;

		confirmed = gs_plugin_ask_untrusted (GS_PLUGIN (data->self), data->title, data->msg, data->details, data->accept_label);

		gs_dnf5_rpm_repo_call_confirm_key (repo_proxy, data->key_id, confirmed, NULL, gs_dnf5_confirm_key_response_cb, NULL);
	}

	return G_SOURCE_REMOVE;
}

typedef struct {
	GsPluginDnf5 *self; /* (not owned) */
	const gchar *session_path; /* (not owned) */
	GsDnf5RpmRepo *repo_proxy; /* (not owned) */
} RepoKeyImportData;

static void
gs_dnf5_repo_key_import_request_cb (GsDnf5Base *object,
				    const gchar *arg_session_object_path,
				    const gchar *arg_key_id,
				    const gchar *const *arg_user_ids,
				    const gchar *arg_key_fingerprint,
				    const gchar *arg_key_url,
				    gint64 arg_timestamp,
				    RepoKeyImportData *repo_key_import_data)
{
	QuestionData *data;
	g_autoptr(GString) user_ids = NULL;
	g_autofree gchar *key_filename = NULL;
	g_autoptr(GString) str_fingerprint = NULL;
	GString *details = NULL;
	const gchar *fingerprint;
	const gchar *from;

	if (g_strcmp0 (repo_key_import_data->session_path, arg_session_object_path) != 0)
		return;

	user_ids = g_string_new (NULL);
	for (guint i = 0; arg_user_ids != NULL && arg_user_ids[i] != NULL; i++) {
		if (i > 0)
			g_string_append (user_ids, ", ");
		g_string_append_c (user_ids, '"');
		g_string_append (user_ids, arg_user_ids[i]);
		g_string_append_c (user_ids, '"');
	}

	g_debug ("%s: key_id:'%s' user_ids:[%s] key_fingerprint:'%s' key_url:'%s' timestamp:%" G_GINT64_FORMAT,
		 G_STRFUNC, arg_key_id, user_ids->str, arg_key_fingerprint, arg_key_url, arg_timestamp);

	if (g_ascii_strncasecmp (arg_key_url, "file:", 5) == 0)
		key_filename = g_filename_from_uri (arg_key_url, NULL, NULL);

	if ((strlen (arg_key_fingerprint) % 4) == 0) {
		str_fingerprint = g_string_new (arg_key_fingerprint);
		for (gint i = str_fingerprint->len - 4; i > 0; i -= 4) {
			g_string_insert_c (str_fingerprint, i, ' ');
		}
	}

	details = g_string_new ("");

	from = key_filename != NULL ? key_filename : arg_key_url;
	fingerprint = str_fingerprint != NULL ? str_fingerprint->str : arg_key_fingerprint;

	#define add_nonempty_line(_format, _value) G_STMT_START { \
		if ((_value) != NULL && *(_value) != '\0') { \
			g_string_append_printf (details, _format, _value); \
			g_string_append_c (details, '\n'); \
		} \
	} G_STMT_END

	/* Translators: the '%s' is replaced with the key user name */
	add_nonempty_line (_("Key user: %s"), user_ids->str);
	/* Translators: the '%s' is replaced with the key fingerprint, a few hex digits */
	add_nonempty_line (_("Fingerprint: %s"), fingerprint);
	/* Translators: the '%s' is replaced with the local path or a URI to the key */
	add_nonempty_line (_("From: %s"), from);

	#undef add_nonempty_line

	data = g_new0 (QuestionData, 1);
	data->title = g_strdup (_("Import Key"));
	/* Translators: the '%s' is replaced with the key ID, usually a few hex digits */
	data->msg = g_strdup_printf (_("Do you want to import key %s?"), arg_key_id);
	data->details = g_string_free (details, FALSE);
	data->accept_label = g_strdup (_("_Import Key"));
	data->key_id = g_strdup (arg_key_id);
	data->self = g_object_ref (repo_key_import_data->self);
	g_weak_ref_init (&data->repo_proxy_weakref, repo_key_import_data->repo_proxy);

	g_idle_add_full (G_PRIORITY_HIGH_IDLE, gs_dnf5_ask_question_idle_cb, data, (GDestroyNotify) question_data_free);
}

static void
gs_dnf5_goal_cancelled_cb (GObject *source_object,
			   GAsyncResult *result,
			   gpointer user_data)
{
	g_autoptr(GCancellable) run_cancellable = user_data;
	g_autoptr(GError) local_error = NULL;
	g_autofree gchar *error_message = NULL;
	gboolean succeeded = FALSE;

	if (gs_dnf5_goal_call_cancel_finish (GS_DNF5_GOAL (source_object), &succeeded, &error_message, result, &local_error)) {
		if (succeeded) {
			g_debug ("%s: Transaction cancelled by the user", G_STRFUNC);
			g_cancellable_cancel (run_cancellable);
		} else {
			g_debug ("%s: Cannot cancel transaction: %s", G_STRFUNC, error_message);
		}
	} else if (!g_error_matches (local_error, G_IO_ERROR, G_IO_ERROR_CANCELLED)) {
		g_debug ("%s: Failed to cancel transaction: %s", G_STRFUNC, local_error->message);
	}
}

static gboolean
gs_dnf5_reset_transaction_sync (GsPluginDnf5 *self,
				const gchar *session_path,
				GsDnf5RpmRpm *rpm_proxy,
				GCancellable *cancellable,
				GError **error)
{
	g_autoptr(GsDnf5Goal) goal_proxy = NULL;

	goal_proxy = gs_dnf5_goal_proxy_new_sync (self->connection,
						  G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						  GS_DNF5_INTERFACE_RPM_DNF,
						  session_path,
						  cancellable,
						  error);
	if (goal_proxy == NULL) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to create Goal proxy: ");
		return FALSE;
	}

	return gs_dnf5_goal_call_reset_sync (goal_proxy, cancellable, error);
}

/* this is needed to receive D-Bus signals about progress in the worker thread */
static gboolean
gs_dnf5_goal_call_do_transaction_sync_wrapper (GsDnf5Goal *proxy,
					       GVariant *arg_options,
					       GCancellable *cancellable,
					       GError **error)
{
	g_autoptr(GMainContext) main_context = g_main_context_ref_thread_default ();
	g_autoptr(GAsyncResult) async_result = NULL;
	gs_dnf5_goal_call_do_transaction (proxy, arg_options, cancellable, gs_dnf5_gather_async_result_cb, &async_result);
	while (!async_result)
		g_main_context_iteration (main_context, TRUE);
	return gs_dnf5_goal_call_do_transaction_finish (proxy, async_result, error);
}

typedef struct {
	GsDnf5Goal *goal_proxy;
	GCancellable *run_cancellable;
} TransactionData;

static void
gs_dnf5_transaction_cancelled_cb (GCancellable *parent_cancellable,
				  TransactionData *transaction_data)
{
	gs_dnf5_goal_call_cancel (transaction_data->goal_proxy, transaction_data->run_cancellable,
				  gs_dnf5_goal_cancelled_cb, g_object_ref (transaction_data->run_cancellable));
}

typedef enum {
	GS_DNF5_TRANSACTION_FLAG_NONE = 0,
	GS_DNF5_TRANSACTION_FLAG_OFFLINE = 1 << 0,
	GS_DNF5_TRANSACTION_FLAG_RESOLVE_ONLY = 1 << 1,
	GS_DNF5_TRANSACTION_FLAG_DISALLOW_ERASING = 1 << 2 /* negative flag, to not have it everywhere */
} GsDnf5TransactionFlags;

static gboolean
gs_dnf5_run_transaction (GsPluginDnf5 *self,
			 const gchar *session_path,
			 GsDnf5RpmRpm *rpm_proxy,
			 GsApp *progress_app,
			 GsAppList *progress_list,
			 GsDnf5TransactionFlags flags,
			 GVariant **out_transaction,
			 GCancellable *cancellable,
			 GError **error)
{
	g_autoptr(GsDnf5Base) base_proxy = NULL;
	g_autoptr(GsDnf5Goal) goal_proxy = NULL;
	g_autoptr(GsDnf5RpmRepo) repo_proxy = NULL;
	g_autoptr(GVariantBuilder) options_builder = NULL;
	g_autoptr(GVariant) transaction = NULL;
	g_autoptr(GsDnf5ProgressHelper) helper = NULL;
	g_autoptr(GCancellable) run_cancellable = NULL;
	RepoKeyImportData repo_key_import_data;
	TransactionData transaction_data;
	gulong repo_key_import_id;
	gulong cancellable_id = 0;
	guint result = 0;
	gboolean success;

	base_proxy = gs_dnf5_base_proxy_new_sync (self->connection,
						  G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						  GS_DNF5_INTERFACE_RPM_DNF,
						  session_path,
						  cancellable,
						  error);
	if (base_proxy == NULL) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to create Base proxy: ");
		return FALSE;
	}

	repo_proxy = gs_dnf5_rpm_repo_proxy_new_sync (self->connection,
						      G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						      GS_DNF5_INTERFACE_RPM_DNF,
						      session_path,
						      cancellable,
						      error);
	if (repo_proxy == NULL) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to create Repo proxy: ");
		return FALSE;
	}

	goal_proxy = gs_dnf5_goal_proxy_new_sync (self->connection,
						  G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						  GS_DNF5_INTERFACE_RPM_DNF,
						  session_path,
						  cancellable,
						  error);
	if (goal_proxy == NULL) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to create Goal proxy: ");
		return FALSE;
	}

	g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (goal_proxy), G_MAXINT);

	options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));

	g_variant_builder_add (options_builder, "{sv}", "allow_erasing",
			       g_variant_new_boolean ((flags & GS_DNF5_TRANSACTION_FLAG_DISALLOW_ERASING) == 0));

	if (progress_app != NULL || progress_list != NULL)
		helper = gs_dnf5_progress_helper_new (base_proxy, rpm_proxy, progress_app, progress_list, session_path);

	repo_key_import_data.self = self;
	repo_key_import_data.session_path = session_path;
	repo_key_import_data.repo_proxy = repo_proxy;
	repo_key_import_id = g_signal_connect (base_proxy, "repo_key_import_request",
					       G_CALLBACK (gs_dnf5_repo_key_import_request_cb), &repo_key_import_data);

	if (!gs_dnf5_goal_call_resolve_sync (goal_proxy,
					     g_variant_builder_end (options_builder),
					     &transaction,
					     &result,
					     cancellable,
					     error)) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to resolve transaction: ");
		g_signal_handler_disconnect (base_proxy, repo_key_import_id);
		return FALSE;
	}

	if (out_transaction != NULL && transaction != NULL)
		*out_transaction = g_variant_ref_sink (transaction);

	/* 0 ... all fine, 1 ... there are warnings, 2 ... failed */
	if (result == 2) {
		g_auto(GStrv) problems = NULL;

		if (gs_dnf5_goal_call_get_transaction_problems_string_sync (goal_proxy, &problems, cancellable, error) &&
		    problems != NULL) {
			/* extra space at the end, but no harm, it reuses translated string */
			GString *msg = g_string_new ("Failed to resolve transaction: ");
			for (guint i = 0; problems[i] != NULL; i++) {
				g_string_append_c (msg, '\n');
				g_string_append (msg, problems[i]);
			}
			g_set_error_literal (error, GS_PLUGIN_ERROR, GS_PLUGIN_ERROR_FAILED, msg->str);
		} else {
			g_set_error_literal (error, GS_PLUGIN_ERROR, GS_PLUGIN_ERROR_FAILED,
				_("Unknown error"));
			g_prefix_error (error, "Failed to resolve transaction: ");
		}
		g_signal_handler_disconnect (base_proxy, repo_key_import_id);
		return FALSE;
	}

	g_clear_pointer (&options_builder, g_variant_builder_unref);

	if ((flags & GS_DNF5_TRANSACTION_FLAG_RESOLVE_ONLY) != 0) {
		g_signal_handler_disconnect (base_proxy, repo_key_import_id);
		return TRUE;
	}

	options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
	if ((flags & GS_DNF5_TRANSACTION_FLAG_OFFLINE) != 0) {
		g_variant_builder_add (options_builder, "{sv}", "offline",
				       g_variant_new_boolean (TRUE));
	}

	run_cancellable = g_cancellable_new ();
	if (cancellable != NULL) {
		transaction_data.goal_proxy = goal_proxy;
		transaction_data.run_cancellable = run_cancellable;

		cancellable_id = g_cancellable_connect (cancellable, G_CALLBACK (gs_dnf5_transaction_cancelled_cb),
							&transaction_data, NULL);
	}

	success = gs_dnf5_goal_call_do_transaction_sync_wrapper (goal_proxy,
								 g_variant_builder_end (options_builder),
								 run_cancellable,
								 error);

	g_cancellable_disconnect (cancellable, cancellable_id);
	g_signal_handler_disconnect (base_proxy, repo_key_import_id);

	/* in case there is a pending async call result for the cancel() invocation */
	g_cancellable_cancel (run_cancellable);

	if (!success) {
		g_auto(GStrv) problems = NULL;

		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to run transaction: ");

		if (error != NULL && *error != NULL &&
		    gs_dnf5_goal_call_get_transaction_problems_string_sync (goal_proxy, &problems, cancellable, NULL) &&
		    problems != NULL) {
			GString *msg = g_string_new ((*error)->message);
			for (guint i = 0; problems[i] != NULL; i++) {
				g_string_append_c (msg, '\n');
				g_string_append (msg, problems[i]);
			}
			g_clear_error (error);
			g_set_error_literal (error, GS_PLUGIN_ERROR, GS_PLUGIN_ERROR_FAILED, msg->str);
		}
	}

	return success;
}

enum OpKindType { OP_KIND_UNKNOWN, OP_KIND_INSTALL, OP_KIND_UPGRADE, OP_KIND_DOWNGRADE, OP_KIND_REINSTALL, OP_KIND_REMOVE, OP_KIND_REPLACED };

static enum OpKindType
gs_dnf5_get_package_op_kind (GVariant *item_array)
{
	struct _op_kinds {
		const gchar *name;
		enum OpKindType kind;
	} op_kinds[] = {
		{ "Install", OP_KIND_INSTALL },
		{ "Upgrade", OP_KIND_UPGRADE },
		{ "Downgrade", OP_KIND_DOWNGRADE },
		{ "Reinstall", OP_KIND_REINSTALL },
		{ "Remove", OP_KIND_REMOVE },
		{ "Replaced", OP_KIND_REPLACED }
	};
	const gchar *str = NULL;

	g_variant_get_child (item_array, 0, "&s", &str);
	if (str == NULL || g_ascii_strcasecmp (str, "package") != 0)
		return OP_KIND_UNKNOWN;

	g_variant_get_child (item_array, 1, "&s", &str);
	if (str == NULL)
		return OP_KIND_UNKNOWN;

	for (guint i = 0; i < G_N_ELEMENTS (op_kinds); i++) {
		if (g_ascii_strcasecmp (str, op_kinds[i].name) == 0) {
			return op_kinds[i].kind;
		}
	}

	return OP_KIND_UNKNOWN;
}

static gboolean
gs_dnf5_gather_update_replaces_cb (GsPluginDnf5 *self,
				   GVariant *item_array,
				   gpointer user_data,
				   GCancellable *cancellable,
				   GError **error)
{
	ReadPackageData *rpd = user_data;
	enum OpKindType op_kind = gs_dnf5_get_package_op_kind (item_array);

	if (op_kind == OP_KIND_REPLACED && rpd->replaces) {
		g_autoptr(GVariantDict) dict = NULL;
		g_autoptr(GVariant) value = NULL;

		value = g_variant_get_child_value (item_array, 4);
		dict = g_variant_dict_new (value);
		g_clear_pointer (&value, g_variant_unref);

		value = g_variant_dict_lookup_value (dict, "id", G_VARIANT_TYPE_INT32);
		if (value != NULL) {
			gint32 id = g_variant_get_int32 (value);
			gchar *version = gs_dnf5_dup_version_from_dict (dict);
			if (version != NULL)
				g_hash_table_insert (rpd->replaces, GINT_TO_POINTER (id), version);
		}
	}

	return TRUE;
}

typedef struct _UpdateReplacesData {
	GsApp *app;
	GHashTable *replaces;
	GHashTable *used_replaces;
} UpdateReplacesData;

static gboolean
gs_dnf5_gather_update_replaced_pkgs_cb (GsPluginDnf5 *self,
					GVariant *item_array,
					gpointer user_data,
					GCancellable *cancellable,
					GError **error)
{
	UpdateReplacesData *upd = user_data;
	gint32 replaces_id = g_variant_get_int32 (item_array);
	const gchar *old_version = g_hash_table_lookup (upd->replaces, GINT_TO_POINTER (replaces_id));
	if (old_version != NULL) {
		gs_app_set_version (upd->app, old_version);
		g_hash_table_add (upd->used_replaces, GINT_TO_POINTER (replaces_id));
	}

	return TRUE;
}

static gboolean
gs_dnf5_gather_updates_from_transaction_cb (GsPluginDnf5 *self,
					    GVariant *item_array,
					    gpointer user_data,
					    GCancellable *cancellable,
					    GError **error)
{
	ReadPackageData *rpd = user_data;
	enum OpKindType op_kind = gs_dnf5_get_package_op_kind (item_array);
	const gchar *str;
	g_autoptr(GVariantDict) dict = NULL;
	g_autoptr(GVariant) value = NULL;
	g_autoptr(GsApp) upd_app = NULL;
	g_autoptr(GString) nevra = NULL;
	GsApp *known_app;

	if (op_kind == OP_KIND_UNKNOWN || (!rpd->cover_unused_replaces && op_kind == OP_KIND_REPLACED) ||
	    (rpd->cover_unused_replaces && op_kind != OP_KIND_REPLACED))
		return TRUE;

	value = g_variant_get_child_value (item_array, 4);
	dict = g_variant_dict_new (value);
	g_clear_pointer (&value, g_variant_unref);

	if (rpd->cover_unused_replaces) {
		value = g_variant_dict_lookup_value (dict, "id", G_VARIANT_TYPE_INT32);
		if (value != NULL) {
			gint32 id = g_variant_get_int32 (value);
			if (g_hash_table_contains (rpd->used_replaces, GINT_TO_POINTER (id)))
				return TRUE;
		} else {
			return TRUE;
		}

		g_clear_pointer (&value, g_variant_unref);
	}

	value = g_variant_dict_lookup_value (dict, "name", G_VARIANT_TYPE_STRING);
	if (value == NULL)
		return TRUE;
	str = g_variant_get_string (value, NULL);
	if (str == NULL)
		return TRUE;

	nevra = g_string_new (str);

	g_clear_pointer (&value, g_variant_unref);
	value = g_variant_dict_lookup_value (dict, "evr", G_VARIANT_TYPE_STRING);
	if (value == NULL)
		return TRUE;
	str = g_variant_get_string (value, NULL);
	if (str == NULL)
		return TRUE;

	g_string_append_c (nevra, '-');
	g_string_append (nevra, str);

	g_clear_pointer (&value, g_variant_unref);
	value = g_variant_dict_lookup_value (dict, "arch", G_VARIANT_TYPE_STRING);
	if (value == NULL)
		return TRUE;
	str = g_variant_get_string (value, NULL);
	if (str == NULL)
		return TRUE;

	g_string_append_c (nevra, '.');
	g_string_append (nevra, str);

	known_app = g_hash_table_lookup (rpd->nevra_to_app, nevra->str);
	if (known_app != NULL) {
		upd_app = g_object_ref (known_app);
		if (gs_app_get_version (upd_app) == NULL)
			gs_dnf5_app_set_version (dict, upd_app, gs_app_set_version);
		else
			gs_dnf5_app_set_version (dict, upd_app, gs_app_set_update_version);

		if (rpd->replaces != NULL) {
			g_autoptr(GVariantDict) dict2 = NULL;

			g_clear_pointer (&value, g_variant_unref);
			value = g_variant_get_child_value (item_array, 3);
			dict2 = g_variant_dict_new (value);
			g_clear_pointer (&value, g_variant_unref);
			value = g_variant_dict_lookup_value (dict2, "replaces", G_VARIANT_TYPE_ARRAY);
			if (value != NULL) {
				UpdateReplacesData upd = { 0, };

				upd.app = upd_app;
				upd.replaces = rpd->replaces;
				upd.used_replaces = rpd->used_replaces;

				if (!gs_dnf5_foreach_item (self, value, gs_dnf5_gather_update_replaced_pkgs_cb, &upd, cancellable, error))
					return FALSE;
			}
		}
	} else {
		upd_app = gs_app_new (NULL);
		gs_app_set_management_plugin (upd_app, GS_PLUGIN (self));
		gs_app_set_metadata (upd_app, "GnomeSoftware::Creator", gs_plugin_get_name (GS_PLUGIN (self)));
		gs_plugin_dnf5_set_packaging_format (upd_app);
		gs_app_set_kind (upd_app, AS_COMPONENT_KIND_GENERIC);
		gs_app_set_bundle_kind (upd_app, AS_BUNDLE_KIND_PACKAGE);
		gs_app_set_scope (upd_app, AS_COMPONENT_SCOPE_SYSTEM);
		gs_app_add_quirk (upd_app, GS_APP_QUIRK_HIDE_EVERYWHERE);
		gs_app_add_quirk (upd_app, GS_APP_QUIRK_NEEDS_REBOOT);

		gs_dnf5_app_set_str (dict, "name", upd_app, gs_app_add_source);
		gs_dnf5_app_set_str2 (dict, "name", upd_app, gs_app_set_name);
		gs_dnf5_app_set_version (dict, upd_app, gs_app_set_version);
		gs_app_add_source_id (upd_app, nevra->str);

		gs_app_list_add (rpd->list, upd_app);
		g_hash_table_insert (rpd->nevra_to_app, g_string_free (g_steal_pointer (&nevra), FALSE), g_object_ref (upd_app));
	}

	gs_dnf5_app_set_size (dict, "install_size", upd_app, gs_app_set_size_installed);
	gs_dnf5_app_set_size (dict, "download_size", upd_app, gs_app_set_size_download);

	switch (op_kind) {
	case OP_KIND_INSTALL:
		gs_app_set_state (upd_app, GS_APP_STATE_AVAILABLE);
		break;
	case OP_KIND_UPGRADE:
	case OP_KIND_DOWNGRADE:
	case OP_KIND_REINSTALL:
		gs_app_set_state (upd_app, GS_APP_STATE_UPDATABLE);
		break;
	case OP_KIND_REMOVE:
	case OP_KIND_REPLACED:
		gs_app_set_state (upd_app, GS_APP_STATE_UNAVAILABLE);
		gs_app_set_size_download (upd_app, GS_SIZE_TYPE_VALID, 0);
		break;
	case OP_KIND_UNKNOWN:
	default:
		g_warn_if_reached ();
		break;
	}

	return TRUE;
}

/* Run in @worker. */
static void
gs_dnf5_list_apps_thread_cb (GTask *task,
			     gpointer source_object,
			     gpointer task_data,
			     GCancellable *cancellable)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	GsAppQueryTristate is_for_update = GS_APP_QUERY_TRISTATE_UNSET;
	const AsComponentKind *component_kinds = NULL;
	GsPluginListAppsData *data = task_data;
	const gchar * const *provides_files = NULL;
	const gchar *provides_tag = NULL;
	GsAppQueryProvidesType provides_type = GS_APP_QUERY_PROVIDES_UNKNOWN;
	g_autofree gchar *session_path = NULL;
	g_autoptr(GsAppList) list = NULL;
	g_autoptr(GsDnf5SessionManager) session_manager = NULL;
	g_autoptr(GsDnf5RpmRpm) rpm_proxy = NULL;
	g_autoptr(GError) local_error = NULL;
	gboolean success = TRUE;

	assert_in_worker (self);

	if (data->query != NULL) {
		is_for_update = gs_app_query_get_is_for_update (data->query);
		component_kinds = gs_app_query_get_component_kinds (data->query);
		provides_files = gs_app_query_get_provides_files (data->query);
		provides_type = gs_app_query_get_provides (data->query, &provides_tag);
	}

	if ((is_for_update == GS_APP_QUERY_TRISTATE_UNSET &&
	     component_kinds == NULL &&
	     provides_tag == NULL &&
	     provides_files == NULL) ||
	    is_for_update == GS_APP_QUERY_TRISTATE_FALSE ||
	    (component_kinds != NULL && !gs_component_kind_array_contains (component_kinds, AS_COMPONENT_KIND_REPOSITORY)) ||
	    gs_app_query_get_n_properties_set (data->query) != 1) {
		g_task_return_new_error (task, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
					 "Unsupported query");
		return;
	}

	session_path = gs_dnf5_open_session (self, GS_DNF5_RELEASEVER_DEFAULT, &session_manager, cancellable, &local_error);
	if (session_path == NULL) {
		gs_dnf5_convert_error (&local_error);
		g_task_return_error (task, g_steal_pointer (&local_error));
		return;
	}

	rpm_proxy = gs_dnf5_rpm_rpm_proxy_new_sync (self->connection,
						    G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						    GS_DNF5_INTERFACE_RPM_DNF,
						    session_path,
						    cancellable,
						    &local_error);
	success = rpm_proxy != NULL;
	if (!success) {
		gs_dnf5_convert_error (&local_error);
		g_prefix_error_literal (&local_error, "Failed to create Rpm proxy: ");
	}

	if (success) {
		g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (rpm_proxy), G_MAXINT);
		list = gs_app_list_new ();

		if (is_for_update == GS_APP_QUERY_TRISTATE_TRUE) {
			g_autoptr(GVariantBuilder) options_builder = NULL;
			g_autoptr(GVariant) result = NULL;

			g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (rpm_proxy), G_MAXINT);

			options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
			g_variant_builder_add (options_builder, "{sv}", "package_attrs",
					       gs_dnf5_dup_package_attrs ());
			g_variant_builder_add (options_builder, "{sv}", "with_provides",
					       g_variant_new_boolean (FALSE));
			g_variant_builder_add (options_builder, "{sv}", "with_filenames",
					       g_variant_new_boolean (FALSE));
			g_variant_builder_add (options_builder, "{sv}", "with_src",
					       g_variant_new_boolean (FALSE));
			g_variant_builder_add (options_builder, "{sv}", "scope",
					       g_variant_new_string ("upgrades"));

			success = gs_dnf5_rpm_rpm_call_list_sync (rpm_proxy,
								  g_variant_builder_end (options_builder),
								  &result,
								  cancellable,
								  &local_error);

			if (success) {
				g_autoptr(GHashTable) nevra_to_app = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
				g_autoptr(GHashTable) replaces = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, g_free);
				g_autoptr(GHashTable) used_replaces = g_hash_table_new (g_direct_hash, g_direct_equal);
				ReadPackageData rpd = { 0, };
				rpd.list = list;
				rpd.set_state = GS_APP_STATE_UPDATABLE;
				rpd.nevra_to_app = nevra_to_app;
				rpd.replaces = replaces;
				rpd.used_replaces = used_replaces;

				success = gs_dnf5_foreach_item (self, result, gs_dnf5_read_package_cb, &rpd, cancellable, &local_error);

				if (success) {
					g_autoptr(GVariantBuilder) options_builder2 = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
					g_autoptr(GVariant) transaction = NULL;
					const gchar * const packages[2] = { NULL, NULL };

					success = gs_dnf5_reset_transaction_sync (self, session_path, rpm_proxy, cancellable, &local_error);

					success = success &&
						  gs_dnf5_rpm_rpm_call_upgrade_sync (rpm_proxy,
										     packages,
										     g_variant_builder_end (options_builder2),
										     cancellable,
										     &local_error);
					success = success &&
						  gs_dnf5_run_transaction (self, session_path, rpm_proxy, NULL, NULL,
									   GS_DNF5_TRANSACTION_FLAG_RESOLVE_ONLY,
									   &transaction, cancellable, &local_error);
					if (success && transaction != NULL) {
						success = gs_dnf5_foreach_item (self, transaction,
										gs_dnf5_gather_update_replaces_cb, &rpd,
										cancellable, &local_error);
						rpd.cover_unused_replaces = FALSE;
						success = success &&
							  gs_dnf5_foreach_item (self, transaction,
										gs_dnf5_gather_updates_from_transaction_cb, &rpd,
										cancellable, &local_error);
						rpd.cover_unused_replaces = TRUE;
						success = success &&
							  gs_dnf5_foreach_item (self, transaction,
										gs_dnf5_gather_updates_from_transaction_cb, &rpd,
										cancellable, &local_error);
					}
				}

				if (success && g_hash_table_size (nevra_to_app) > 0) {
					g_autoptr(GsDnf5Advisory) advisory_proxy = NULL;
					g_autoptr(GError) local_error2 = NULL;
					gboolean adv_success;

					advisory_proxy = gs_dnf5_advisory_proxy_new_sync (self->connection,
											  G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
											  GS_DNF5_INTERFACE_RPM_DNF,
											  session_path,
											  cancellable,
											  &local_error2);
					adv_success = advisory_proxy != NULL;

					if (adv_success) {
						g_autoptr(GVariantBuilder) adv_options_builder = NULL;
						g_autoptr(GVariantBuilder) attrs_builder = NULL;
						g_autoptr(GVariantBuilder) severities_builder = NULL;
						g_autoptr(GVariant) adv_result = NULL;

						attrs_builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
						g_variant_builder_add (attrs_builder, "s", "severity");
						g_variant_builder_add (attrs_builder, "s", "collections");

						severities_builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
						g_variant_builder_add (severities_builder, "s", "critical");

						adv_options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
						g_variant_builder_add (adv_options_builder, "{sv}", "advisory_attrs",
								       g_variant_builder_end (attrs_builder));
						g_variant_builder_add (adv_options_builder, "{sv}", "availability",
								       g_variant_new_string ("updates"));
						g_variant_builder_add (adv_options_builder, "{sv}", "severities",
								       g_variant_builder_end (severities_builder));

						adv_success = gs_dnf5_advisory_call_list_sync (advisory_proxy,
											       g_variant_builder_end (adv_options_builder),
											       &adv_result,
											       cancellable,
											       &local_error2);
						if (adv_success) {
							adv_success = gs_dnf5_foreach_item (self, adv_result,
											    gs_dnf5_refine_from_advisory_cb, nevra_to_app,
											    cancellable, &local_error2);
						} else {
							g_debug ("Failed to call Advisory::list: %s", local_error2->message);
						}
					} else {
						g_debug ("Failed to create Advisory proxy: %s", local_error2->message);
					}
				}
			} else {
				gs_dnf5_convert_error (&local_error);
				g_prefix_error_literal (&local_error, "Failed to call Rpm::list: ");
			}
		} else if (gs_component_kind_array_contains (component_kinds, AS_COMPONENT_KIND_REPOSITORY)) {
			g_autoptr(GsDnf5RpmRepo) repo_proxy = NULL;

			repo_proxy = gs_dnf5_rpm_repo_proxy_new_sync (self->connection,
								      G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
								      GS_DNF5_INTERFACE_RPM_DNF,
								      session_path,
								      cancellable,
								      &local_error);
			success = repo_proxy != NULL;
			if (success) {
				g_autoptr(GVariantBuilder) options_builder = NULL;
				g_autoptr(GVariantBuilder) attrs_builder = NULL;
				g_autoptr(GVariant) result = NULL;

				g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (repo_proxy), G_MAXINT);

				attrs_builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
				g_variant_builder_add (attrs_builder, "s", "id");
				g_variant_builder_add (attrs_builder, "s", "name");
				g_variant_builder_add (attrs_builder, "s", "enabled");

				options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
				g_variant_builder_add (options_builder, "{sv}", "repo_attrs",
						       g_variant_builder_end (attrs_builder));
				g_variant_builder_add (options_builder, "{sv}", "enable_disable",
						       g_variant_new_string ("all"));

				success = gs_dnf5_rpm_repo_call_list_sync (repo_proxy,
									   g_variant_builder_end (options_builder),
									   &result,
									   cancellable,
									   &local_error);

				if (success) {
					success = gs_dnf5_foreach_item (self, result, gs_dnf5_read_repo_cb, list, cancellable, &local_error);
				} else {
					gs_dnf5_convert_error (&local_error);
					g_prefix_error_literal (&local_error, "Failed to call list: ");
				}
			} else {
				gs_dnf5_convert_error (&local_error);
				g_prefix_error_literal (&local_error, "Failed to create Rpm repo proxy: ");
			}
		} else if (provides_files != NULL) {
			g_autoptr(GVariantBuilder) patterns_builder = NULL;
			g_autoptr(GVariantBuilder) options_builder = NULL;
			g_autoptr(GVariant) result = NULL;

			patterns_builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
			for (guint i = 0; provides_files[i] != NULL; i++) {
				g_variant_builder_add (patterns_builder, "s", provides_files[i]);
			}

			options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
			g_variant_builder_add (options_builder, "{sv}", "scope",
					       g_variant_new_string ("all"));
			g_variant_builder_add (options_builder, "{sv}", "with_nevra",
					       g_variant_new_boolean (FALSE));
			g_variant_builder_add (options_builder, "{sv}", "with_provides",
					       g_variant_new_boolean (TRUE));
			g_variant_builder_add (options_builder, "{sv}", "with_filenames",
					       g_variant_new_boolean (TRUE));
			g_variant_builder_add (options_builder, "{sv}", "with_src",
					       g_variant_new_boolean (FALSE));
			g_variant_builder_add (options_builder, "{sv}", "package_attrs",
					       gs_dnf5_dup_package_attrs ());
			g_variant_builder_add (options_builder, "{sv}", "patterns",
					       g_variant_builder_end (patterns_builder));

			success = gs_dnf5_rpm_rpm_call_list_sync (rpm_proxy,
								  g_variant_builder_end (options_builder),
								  &result,
								  cancellable,
								  &local_error);

			if (success) {
				ReadPackageData rpd = { 0, };
				rpd.list = list;
				rpd.set_state = GS_APP_STATE_UNKNOWN;
				rpd.nevra_to_app = NULL;
				success = gs_dnf5_foreach_item (self, result, gs_dnf5_read_package_cb, &rpd, cancellable, &local_error);
			} else {
				gs_dnf5_convert_error (&local_error);
				g_prefix_error_literal (&local_error, "Failed to call list: ");
			}
		} else if (provides_tag != NULL) {
			g_autoptr(GVariantBuilder) patterns_builder = NULL;
			g_autoptr(GVariantBuilder) options_builder = NULL;
			g_autoptr(GVariant) result = NULL;
			g_autofree gchar *pattern1 = NULL, *pattern2 = NULL;

			patterns_builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));

			switch (provides_type) {
			case GS_APP_QUERY_PROVIDES_PACKAGE_NAME:
				g_variant_builder_add (patterns_builder, "s", provides_tag);
				break;
			case GS_APP_QUERY_PROVIDES_GSTREAMER:
				pattern1 = g_strdup_printf ("gstreamer0.10(%s)", provides_tag);
				pattern2 = g_strdup_printf ("gstreamer1(%s)", provides_tag);
				break;
			case GS_APP_QUERY_PROVIDES_FONT:
				pattern1 = g_strdup_printf ("font(%s)", provides_tag);
				break;
			case GS_APP_QUERY_PROVIDES_MIME_HANDLER:
				pattern1 = g_strdup_printf ("mimehandler(%s)", provides_tag);
				break;
			case GS_APP_QUERY_PROVIDES_PS_DRIVER:
				pattern1 = g_strdup_printf ("postscriptdriver(%s)", provides_tag);
				break;
			case GS_APP_QUERY_PROVIDES_PLASMA:
				pattern1 = g_strdup_printf ("plasma4(%s)", provides_tag);
				pattern2 = g_strdup_printf ("plasma5(%s)", provides_tag);
				break;
			case GS_APP_QUERY_PROVIDES_UNKNOWN:
			default:
				g_assert_not_reached ();
			}
			if (pattern1 != NULL)
				g_variant_builder_add (patterns_builder, "s", pattern1);
			if (pattern2 != NULL)
				g_variant_builder_add (patterns_builder, "s", pattern2);

			options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
			g_variant_builder_add (options_builder, "{sv}", "scope",
					       g_variant_new_string ("all"));
			g_variant_builder_add (options_builder, "{sv}", "with_nevra",
					       g_variant_new_boolean (provides_type == GS_APP_QUERY_PROVIDES_PACKAGE_NAME));
			g_variant_builder_add (options_builder, "{sv}", "with_provides",
					       g_variant_new_boolean (provides_type != GS_APP_QUERY_PROVIDES_PACKAGE_NAME));
			g_variant_builder_add (options_builder, "{sv}", "with_filenames",
					       g_variant_new_boolean (FALSE));
			g_variant_builder_add (options_builder, "{sv}", "with_src",
					       g_variant_new_boolean (FALSE));
			g_variant_builder_add (options_builder, "{sv}", "package_attrs",
					       gs_dnf5_dup_package_attrs ());
			g_variant_builder_add (options_builder, "{sv}", "patterns",
					       g_variant_builder_end (patterns_builder));

			success = gs_dnf5_rpm_rpm_call_list_sync (rpm_proxy,
								  g_variant_builder_end (options_builder),
								  &result,
								  cancellable,
								  &local_error);

			if (success) {
				ReadPackageData rpd = { 0, };
				rpd.list = list;
				rpd.set_state = GS_APP_STATE_UNKNOWN;
				rpd.nevra_to_app = NULL;
				success = gs_dnf5_foreach_item (self, result, gs_dnf5_read_package_cb, &rpd, cancellable, &local_error);
			} else {
				gs_dnf5_convert_error (&local_error);
				g_prefix_error_literal (&local_error, "Failed to call list: ");
			}
		} else {
			g_assert_not_reached ();
		}
	}

	gs_dnf5_close_session (self, session_manager, session_path);

	if (success) {
		g_task_return_pointer (task, g_steal_pointer (&list), g_object_unref);
	} else {
		gs_dnf5_convert_error (&local_error);
		g_task_return_error (task, g_steal_pointer (&local_error));
	}
}

static void
gs_plugin_dnf5_list_apps_async (GsPlugin *plugin,
				GsAppQuery *query,
				GsPluginListAppsFlags flags,
				GsPluginEventCallback event_callback,
				void *event_user_data,
				GCancellable *cancellable,
				GAsyncReadyCallback callback,
				gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	gboolean interactive = (flags & GS_PLUGIN_LIST_APPS_FLAGS_INTERACTIVE) != 0;
	g_autoptr(GTask) task = NULL;

	task = gs_plugin_list_apps_data_new_task (plugin, query, flags, event_callback, event_user_data,
						  cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_list_apps_async);

	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_list_apps_thread_cb, g_steal_pointer (&task));
}

static GsAppList *
gs_plugin_dnf5_list_apps_finish (GsPlugin *plugin,
				 GAsyncResult *result,
				 GError **error)
{
	return g_task_propagate_pointer (G_TASK (result), error);
}

static gboolean
gs_dnf5_gather_dependency_sizes_cb (GsPluginDnf5 *self,
				    GVariant *item_array,
				    gpointer user_data,
				    GCancellable *cancellable,
				    GError **error)
{
	GsApp *main_app = user_data;
	const gchar *str;
	g_autoptr(GVariantDict) dict = NULL;
	g_autoptr(GVariant) value = NULL;
	g_autoptr(GsApp) dep_app = NULL;

	g_variant_get_child (item_array, 0, "&s", &str);
	if (str == NULL || g_ascii_strcasecmp (str, "package") != 0)
		return TRUE;

	g_variant_get_child (item_array, 2, "&s", &str);
	if (str == NULL || g_ascii_strcasecmp (str, "dependency") != 0)
		return TRUE;

	value = g_variant_get_child_value (item_array, 4);
	dict = g_variant_dict_new (value);
	g_clear_pointer (&value, g_variant_unref);

	value = g_variant_dict_lookup_value (dict, "name", G_VARIANT_TYPE_STRING);
	if (value == NULL)
		return TRUE;
	str = g_variant_get_string (value, NULL);
	if (str == NULL)
		return TRUE;
	/* the app itself is in the list too */
	if (g_strcmp0 (gs_app_get_default_source (main_app), str) == 0)
		return TRUE;

	/* always create a new app, no cached can be used here */
	dep_app = gs_app_new (NULL);
	gs_app_set_management_plugin (dep_app, GS_PLUGIN (self));
	gs_app_set_metadata (dep_app, "GnomeSoftware::Creator", gs_plugin_get_name (GS_PLUGIN (self)));
	gs_plugin_dnf5_set_packaging_format (dep_app);
	gs_app_set_kind (dep_app, AS_COMPONENT_KIND_GENERIC);
	gs_app_set_bundle_kind (dep_app, AS_BUNDLE_KIND_PACKAGE);
	gs_app_set_scope (dep_app, AS_COMPONENT_SCOPE_SYSTEM);
	gs_app_add_quirk (dep_app, GS_APP_QUIRK_HIDE_EVERYWHERE);

	gs_dnf5_app_set_str (dict, "name", dep_app, gs_app_add_source);
	gs_dnf5_app_set_str (dict, "nevra", dep_app, gs_app_add_source_id);
	gs_dnf5_app_set_str2 (dict, "name", dep_app, gs_app_set_name);
	gs_dnf5_app_set_str2 (dict, "summary", dep_app, gs_app_set_summary);
	gs_dnf5_app_set_str2 (dict, "description", dep_app, gs_app_set_description);
	gs_dnf5_app_set_str2 (dict, "license", dep_app, gs_app_set_license);
	gs_dnf5_app_set_size (dict, "install_size", dep_app, gs_app_set_size_installed);
	gs_dnf5_app_set_size (dict, "download_size", dep_app, gs_app_set_size_download);

	gs_dnf5_update_app_state (dep_app, GS_APP_STATE_UNKNOWN, dict);

	gs_app_add_related (main_app, dep_app);

	return TRUE;
}

static gpointer
gs_dnf5_dependency_sizes_thread (gpointer thread_data)
{
	g_autofree GWeakRef *weakref = thread_data;
	g_autoptr(GsPluginDnf5) self = g_weak_ref_get (weakref);
	g_autoptr(GCancellable) cancellable = NULL;
	g_autoptr(GError) local_error = NULL;
	guint index = 0;

	if (self == NULL)
		return NULL;

	g_mutex_lock (&self->dependency_sizes_data.mutex);

	if (self->dependency_sizes_data.cancellable != NULL)
		cancellable = g_object_ref (self->dependency_sizes_data.cancellable);

	while (self->dependency_sizes_data.apps != NULL && index < self->dependency_sizes_data.apps->len &&
	       !g_cancellable_is_cancelled (self->dependency_sizes_data.cancellable)) {
		GsApp *app = g_ptr_array_index (self->dependency_sizes_data.apps, index);
		g_autofree gchar *session_path = NULL;
		g_autoptr(GsDnf5SessionManager) session_manager = NULL;
		g_autoptr(GsDnf5RpmRpm) rpm_proxy = NULL;
		gboolean success;

		index++;

		if (gs_app_is_installed (app) ||
		    gs_app_list_length (gs_app_get_related (app)) > 0)
			continue;

		g_mutex_unlock (&self->dependency_sizes_data.mutex);

		/* need to use temporary session, because the goals are piled up and cannot be cancelled yet */
		session_path = gs_dnf5_open_session (self, GS_DNF5_RELEASEVER_DEFAULT, &session_manager, cancellable, &local_error);

		success = session_path != NULL;
		if (!success)
			gs_dnf5_convert_error (&local_error);

		if (success) {
			rpm_proxy = gs_dnf5_rpm_rpm_proxy_new_sync (self->connection,
								    G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
								    GS_DNF5_INTERFACE_RPM_DNF,
								    session_path,
								    cancellable,
								    &local_error);
			success = rpm_proxy != NULL;
			if (!success) {
				gs_dnf5_convert_error (&local_error);
				g_prefix_error_literal (&local_error, "Failed to create Rpm proxy: ");
			}
		}

		if (success) {
			g_autoptr(GVariantBuilder) options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
			g_autoptr(GVariant) transaction = NULL;
			const gchar *package_name = gs_app_get_default_source (app);
			const gchar *packages[2] = { NULL, NULL };

			packages[0] = package_name;

			success = gs_dnf5_reset_transaction_sync (self, session_path, rpm_proxy, cancellable, &local_error);

			success = success &&
				  gs_dnf5_rpm_rpm_call_install_sync (rpm_proxy,
								     (const gchar * const *) packages,
								     g_variant_builder_end (options_builder),
								     cancellable,
								     &local_error);
			success = success &&
				  gs_dnf5_run_transaction (self, session_path, rpm_proxy, NULL, NULL,
							   GS_DNF5_TRANSACTION_FLAG_RESOLVE_ONLY,
							   &transaction, cancellable, &local_error);
			if (success && transaction != NULL)
				success = gs_dnf5_foreach_item (self, transaction, gs_dnf5_gather_dependency_sizes_cb, app, cancellable, &local_error);
		}

		gs_dnf5_close_session (self, session_manager, session_path);

		g_mutex_lock (&self->dependency_sizes_data.mutex);

		if (!success)
			break;
	}

	g_clear_pointer (&self->dependency_sizes_data.apps, g_ptr_array_unref);
	g_clear_object (&self->dependency_sizes_data.cancellable);
	g_mutex_unlock (&self->dependency_sizes_data.mutex);

	if (local_error != NULL)
		g_debug ("Failed to get app dependency sizes: %s", local_error->message);

	return NULL;
}

static gboolean
gs_dnf5_get_dependency_sizes_sync (GsPluginDnf5 *self,
				   GHashTable *apps,
				   GCancellable *cancellable,
				   GError **error)
{
	g_autoptr(GMutexLocker) locker = g_mutex_locker_new (&self->session_data.mutex);
	GHashTableIter iter;
	gpointer value = NULL;
	gboolean need_thread = FALSE;

	g_hash_table_iter_init (&iter, apps);
	while (g_hash_table_iter_next (&iter, NULL, &value)) {
		GsApp *app = value;
		GsAppList *related = gs_app_get_related (app);

		if (gs_app_has_quirk (app, GS_APP_QUIRK_HIDE_EVERYWHERE) ||
		    gs_app_get_state (app) == GS_APP_STATE_AVAILABLE_LOCAL ||
		    gs_app_is_installed (app) ||
		    gs_app_list_length (related) > 0)
			continue;

		/* getting list of the dependencies to be installed can take a long time, thus run it
		   in a new thread, to not block the refine job */
		if (self->dependency_sizes_data.apps == NULL) {
			need_thread = TRUE;
			self->dependency_sizes_data.apps = g_ptr_array_new_with_free_func (g_object_unref);
		}

		if (!g_ptr_array_find (self->dependency_sizes_data.apps, app, NULL))
			g_ptr_array_add (self->dependency_sizes_data.apps, g_object_ref (app));
	}

	if (need_thread) {
		GWeakRef *weakref = g_new0 (GWeakRef, 1);

		g_assert (self->dependency_sizes_data.cancellable == NULL);
		self->dependency_sizes_data.cancellable = g_cancellable_new ();

		g_weak_ref_set (weakref, self);

		g_thread_unref (g_thread_new ("gs-dnf5-dependency-sizes", gs_dnf5_dependency_sizes_thread, g_steal_pointer (&weakref)));
	}

	return TRUE;
}

static gboolean
gs_dnf5_refine_cb (GsPluginDnf5 *self,
		   GVariant *package_array,
		   gpointer user_data,
		   GCancellable *cancellable,
		   GError **error)
{
	GHashTable *apps = user_data;
	const gchar *name;
	GsApp *app;
	g_autoptr(GVariantDict) dict = NULL;
	g_autoptr(GVariant) value = NULL;

	dict = g_variant_dict_new (package_array);

	value = g_variant_dict_lookup_value (dict, "name", G_VARIANT_TYPE_STRING);
	if (value == NULL)
		return TRUE;
	name = g_variant_get_string (value, NULL);
	if (name == NULL)
		return TRUE;
	app = g_hash_table_lookup (apps, name);
	if (app == NULL)
		return TRUE;

	if (gs_app_has_management_plugin (app, NULL) &&
	    gs_app_get_bundle_kind (app) == AS_BUNDLE_KIND_PACKAGE &&
	    gs_app_get_scope (app) == AS_COMPONENT_SCOPE_SYSTEM) {
		gs_app_set_management_plugin (app, GS_PLUGIN (self));
		gs_plugin_dnf5_set_packaging_format (app);
	}

	gs_dnf5_app_set_version (dict, app, gs_app_set_version);
	gs_dnf5_app_set_size (dict, "install_size", app, gs_app_set_size_installed);
	gs_dnf5_app_set_size (dict, "download_size", app, gs_app_set_size_download);

	gs_dnf5_update_app_state (app, GS_APP_STATE_UNKNOWN, dict);

	return TRUE;
}

/* Run in @worker. */
static void
gs_dnf5_refine_thread_cb (GTask *task,
			  gpointer source_object,
			  gpointer task_data,
			  GCancellable *cancellable)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	GsPlugin *plugin = GS_PLUGIN (self);
	GsPluginRefineData *data = task_data;
	g_autofree gchar *session_path = NULL;
	g_autoptr(GHashTable) apps = NULL;
	g_autoptr(GsDnf5SessionManager) session_manager = NULL;
	g_autoptr(GsDnf5RpmRpm) rpm_proxy = NULL;
	g_autoptr(GError) local_error = NULL;
	gboolean success = TRUE;

	assert_in_worker (self);

	apps = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
	for (guint i = 0; i < gs_app_list_length (data->list); i++) {
		GsApp *app = gs_app_list_index (data->list, i);
		if (!gs_app_has_quirk (app, GS_APP_QUIRK_IS_WILDCARD) && (
		    gs_app_get_state (app) == GS_APP_STATE_UNKNOWN ||
		    ((data->require_flags & GS_PLUGIN_REFINE_REQUIRE_FLAGS_SIZE_DATA) != 0 &&
		      gs_app_list_length (gs_app_get_related (app)) == 0))) {
			if (gs_app_get_default_source (app) != NULL &&
			    ((gs_app_has_management_plugin (app, NULL) &&
			      gs_app_get_bundle_kind (app) == AS_BUNDLE_KIND_PACKAGE &&
			      gs_app_get_scope (app) == AS_COMPONENT_SCOPE_SYSTEM) ||
			     gs_app_has_management_plugin (app, plugin)))
				g_hash_table_insert (apps, g_strdup (gs_app_get_default_source (app)), g_object_ref (app));
		}
	}

	if (g_hash_table_size (apps) == 0) {
		g_task_return_boolean (task, TRUE);
		return;
	}

	session_path = gs_dnf5_open_session (self, GS_DNF5_RELEASEVER_DEFAULT, &session_manager, cancellable, &local_error);
	if (session_path == NULL) {
		gs_dnf5_convert_error (&local_error);
		g_task_return_error (task, g_steal_pointer (&local_error));
		return;
	}

	rpm_proxy = gs_dnf5_rpm_rpm_proxy_new_sync (self->connection,
						    G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						    GS_DNF5_INTERFACE_RPM_DNF,
						    session_path,
						    cancellable,
						    &local_error);
	success = rpm_proxy != NULL;
	if (!success) {
		gs_dnf5_convert_error (&local_error);
		g_prefix_error_literal (&local_error, "Failed to create Rpm proxy: ");
	}

	if (success) {
		g_autoptr(GVariantBuilder) attrs_builder = NULL;
		g_autoptr(GVariantBuilder) options_builder = NULL;
		g_autoptr(GVariantBuilder) patterns_builder = NULL;
		g_autoptr(GVariant) result = NULL;
		GHashTableIter iter;
		gpointer key = NULL;

		g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (rpm_proxy), G_MAXINT);

		patterns_builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
		g_hash_table_iter_init (&iter, apps);
		while (g_hash_table_iter_next (&iter, &key, NULL)) {
			const gchar *name = key;
			g_variant_builder_add (patterns_builder, "s", name);
		}

		attrs_builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
		g_variant_builder_add (attrs_builder, "s", "name");
		g_variant_builder_add (attrs_builder, "s", "version");
		g_variant_builder_add (attrs_builder, "s", "install_size");
		g_variant_builder_add (attrs_builder, "s", "download_size");
		g_variant_builder_add (attrs_builder, "s", "is_installed");

		options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
		g_variant_builder_add (options_builder, "{sv}", "patterns",
				       g_variant_builder_end (patterns_builder));
		g_variant_builder_add (options_builder, "{sv}", "package_attrs",
				       g_variant_builder_end (attrs_builder));
		g_variant_builder_add (options_builder, "{sv}", "with_nevra",
				       g_variant_new_boolean (TRUE));
		g_variant_builder_add (options_builder, "{sv}", "with_provides",
				       g_variant_new_boolean (FALSE));
		g_variant_builder_add (options_builder, "{sv}", "with_filenames",
				       g_variant_new_boolean (FALSE));
		g_variant_builder_add (options_builder, "{sv}", "with_src",
				       g_variant_new_boolean (FALSE));
		g_variant_builder_add (options_builder, "{sv}", "scope",
				       g_variant_new_string ("all"));
		g_variant_builder_add (options_builder, "{sv}", "latest_limit",
				       g_variant_new_int32 (1));

		success = gs_dnf5_rpm_rpm_call_list_sync (rpm_proxy,
							  g_variant_builder_end (options_builder),
							  &result,
							  cancellable,
							  &local_error);

		if (success) {
			success = gs_dnf5_foreach_item (self, result, gs_dnf5_refine_cb, apps, cancellable, &local_error);

			if (success &&
			    (data->require_flags & GS_PLUGIN_REFINE_REQUIRE_FLAGS_SIZE_DATA) != 0) {
				/* do not panic when cannot get dependency sizes for the apps */
				if (!gs_dnf5_get_dependency_sizes_sync (self, apps, cancellable, &local_error)) {
					g_debug ("Failed to get dependency sizes: %s", local_error->message);
					g_clear_error (&local_error);
				}
			}
		} else {
			gs_dnf5_convert_error (&local_error);
			g_prefix_error_literal (&local_error, "Failed to call list: ");
		}
	}

	gs_dnf5_close_session (self, session_manager, session_path);

	if (success) {
		g_task_return_boolean (task, TRUE);
	} else {
		gs_dnf5_convert_error (&local_error);
		g_task_return_error (task, g_steal_pointer (&local_error));
	}
}

static void
gs_plugin_dnf5_refine_async (GsPlugin *plugin,
			     GsAppList *list,
			     GsPluginRefineFlags job_flags,
			     GsPluginRefineRequireFlags require_flags,
			     GsPluginEventCallback event_callback,
			     void *event_user_data,
			     GCancellable *cancellable,
			     GAsyncReadyCallback callback,
			     gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	gboolean interactive = (job_flags & GS_PLUGIN_REFINE_FLAGS_INTERACTIVE) != 0;
	g_autoptr(GTask) task = NULL;

	task = gs_plugin_refine_data_new_task (plugin, list, job_flags, require_flags,
					       event_callback, event_user_data,
					       cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_refine_async);

	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_refine_thread_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_refine_finish (GsPlugin *plugin,
			      GAsyncResult *result,
			      GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

/* Run in @worker. */
static void
gs_dnf5_refresh_metadata_thread_cb (GTask *task,
				    gpointer source_object,
				    gpointer task_data,
				    GCancellable *cancellable)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	g_autofree gchar *session_path = NULL;
	g_autoptr(GsDnf5SessionManager) session_manager = NULL;
	g_autoptr(GsDnf5Base) base_proxy = NULL;
	g_autoptr(GError) local_error = NULL;
	gboolean success = TRUE;

	assert_in_worker (self);

	session_path = gs_dnf5_open_session (self, GS_DNF5_RELEASEVER_DEFAULT, &session_manager, cancellable, &local_error);
	if (session_path == NULL) {
		gs_dnf5_convert_error (&local_error);
		g_task_return_error (task, g_steal_pointer (&local_error));
		return;
	}

	base_proxy = gs_dnf5_base_proxy_new_sync (self->connection,
						  G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						  GS_DNF5_INTERFACE_RPM_DNF,
						  session_path,
						  cancellable,
						  &local_error);
	success = base_proxy != NULL;
	if (!success) {
		gs_dnf5_convert_error (&local_error);
		g_prefix_error_literal (&local_error, "Failed to create Base proxy: ");
	}

	if (success && GPOINTER_TO_INT (task_data) == 1) {
		g_autofree gchar *op_error_msg = NULL;
		gboolean op_success = FALSE;

		success = gs_dnf5_base_call_clean_sync (base_proxy,
							"expire-cache",
							&op_success,
							&op_error_msg,
							cancellable,
							&local_error);
		if (!success)
			g_debug ("Failed to call expire-cache: %s", local_error->message);
		else if (!op_success)
			g_debug ("Failed to expire-cache: %s", op_error_msg);

		/* ignore expire-cache error, only log about it */
		g_clear_error (&local_error);

		/* reset the session, to take the expiration into the effect */
		success = gs_dnf5_reset_session_sync (self, session_path, cancellable, &local_error);
	}

	if (success) {
		gboolean result = FALSE;

		g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (base_proxy), G_MAXINT);

		success = gs_dnf5_base_call_read_all_repos_sync (base_proxy, &result, cancellable, &local_error);

		/* 'result' is ignored here, due to no error provided */
		if (!success) {
			gs_dnf5_convert_error (&local_error);
			g_prefix_error_literal (&local_error, "Failed to call read-all-repos: ");
		}
	}

	gs_dnf5_close_session (self, session_manager, session_path);

	if (success) {
		g_task_return_boolean (task, TRUE);
	} else {
		gs_dnf5_convert_error (&local_error);
		g_task_return_error (task, g_steal_pointer (&local_error));
	}
}

static void
gs_plugin_dnf5_refresh_metadata_async (GsPlugin *plugin,
				       guint64 cache_age_secs,
				       GsPluginRefreshMetadataFlags flags,
				       GsPluginEventCallback event_callback,
				       void *event_user_data,
				       GCancellable *cancellable,
				       GAsyncReadyCallback callback,
				       gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	gboolean interactive = (flags & GS_PLUGIN_REFRESH_METADATA_FLAGS_INTERACTIVE) != 0;
	g_autoptr(GTask) task = NULL;

	task = g_task_new (self, cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_refresh_metadata_async);
	/* no need to save whole structure, the seconds are not used at all,
	   thus make it only a flag whether to invalidate the cache or not first */
	g_task_set_task_data (task, GINT_TO_POINTER (cache_age_secs < 60 ? 1 : 0), NULL);

	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_refresh_metadata_thread_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_refresh_metadata_finish (GsPlugin *plugin,
					GAsyncResult *result,
					GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

void
gs_plugin_adopt_app (GsPlugin *plugin,
		     GsApp *app)
{
	if (gs_app_get_bundle_kind (app) == AS_BUNDLE_KIND_PACKAGE &&
	    gs_app_get_scope (app) == AS_COMPONENT_SCOPE_SYSTEM) {
		gs_app_set_management_plugin (app, plugin);
		gs_plugin_dnf5_set_packaging_format (app);
		return;
	} else if (gs_app_get_kind (app) == AS_COMPONENT_KIND_OPERATING_SYSTEM) {
		gs_app_set_management_plugin (app, plugin);
	}
}

static gboolean
gs_dnf5_install_update (GsPlugin *plugin,
			GsAppList *list,
			gboolean is_install,
			gboolean interactive,
			GsPluginEventCallback event_callback,
			void *event_user_data,
			GCancellable *cancellable,
			GError **error)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	g_autofree gchar *session_path = NULL;
	g_autoptr(GPtrArray) packages = NULL;
	g_autoptr(GsAppList) used_apps = NULL;
	g_autoptr(GsDnf5SessionManager) session_manager = NULL;
	g_autoptr(GsDnf5RpmRpm) rpm_proxy = NULL;
	g_autoptr(GsApp) progress_app = NULL;
	GsAppList *progress_list = is_install ? NULL : list;
	gboolean success;

	/* queue for install if installation needs the network */
	if (!gs_plugin_get_network_available (plugin)) {
		if (is_install) {
			for (guint i = 0; i < gs_app_list_length (list); i++) {
				GsApp *app = gs_app_list_index (list, i);
				if (gs_app_has_management_plugin (app, plugin)) {
					if (gs_app_get_kind (app) == AS_COMPONENT_KIND_REPOSITORY) {
						g_warn_if_reached ();
						continue;
					}

					gs_app_set_state (app, GS_APP_STATE_QUEUED_FOR_INSTALL);
				}
			}
		}
		return TRUE;
	}

	packages = g_ptr_array_new_with_free_func (g_free);
	used_apps = gs_app_list_new ();

	for (guint i = 0; i < gs_app_list_length (list); i++) {
		GsApp *app = gs_app_list_index (list, i);
		if (gs_app_has_management_plugin (app, plugin)) {
			const gchar *source = gs_app_get_default_source (app);
			if (source != NULL) {
				GFile *local_file = gs_app_get_local_file (app);
				if (local_file != NULL)
					g_ptr_array_add (packages, g_file_get_uri (local_file));
				else
					g_ptr_array_add (packages, g_strdup (source));
				gs_app_list_add (used_apps, app);
				if (progress_list == NULL && progress_app == NULL)
					progress_app = g_object_ref (app);
			}
		} else if (gs_app_has_quirk (app, GS_APP_QUIRK_IS_PROXY)) {
			GsAppList *related = gs_app_get_related (app);
			gboolean is_os_update = gs_app_get_special_kind (app) == GS_APP_SPECIAL_KIND_OS_UPDATE;
			for (guint j = 0; j < gs_app_list_length (related); j++) {
				GsApp *rel_app = gs_app_list_index (related, j);
				/* add only updatable packages into the list when it's an OS Update, otherwise
				   dnf5 claims "Packages for argument '$pkgname' available, but not installed." */
				if (gs_app_has_management_plugin (rel_app, plugin) && (!is_install || !is_os_update ||
				    gs_app_get_state (rel_app) == GS_APP_STATE_UPDATABLE)) {
					const gchar *source = gs_app_get_default_source (rel_app);
					if (source != NULL) {
						GFile *local_file = gs_app_get_local_file (rel_app);
						if (local_file != NULL)
							g_ptr_array_add (packages, g_file_get_uri (local_file));
						else
							g_ptr_array_add (packages, g_strdup (source));
						gs_app_list_add (used_apps, rel_app);
					}
				}
			}
			if (progress_list == NULL && progress_app == NULL)
				progress_app = g_object_ref (app);
		}
	}

	if (packages->len == 0) {
		g_set_error_literal (error,
				     GS_PLUGIN_ERROR,
				     GS_PLUGIN_ERROR_NOT_SUPPORTED,
				     "installing not available");
		return FALSE;
	}

	session_path = gs_dnf5_open_session (self, GS_DNF5_RELEASEVER_DEFAULT, &session_manager, cancellable, error);
	if (session_path == NULL) {
		gs_dnf5_convert_error (error);
		return FALSE;
	}

	for (guint i = 0; i < gs_app_list_length (used_apps); i++) {
		GsApp *app = gs_app_list_index (used_apps, i);
		g_autoptr(GsAppList) addons = NULL;

		if (gs_app_get_kind (app) == AS_COMPONENT_KIND_REPOSITORY) {
			g_warn_if_reached ();
			continue;
		}

		if (is_install) {
			gs_app_set_state (app, GS_APP_STATE_INSTALLING);
			addons = gs_app_dup_addons (app);
		}
		for (guint j = 0; addons != NULL && j < gs_app_list_length (addons); j++) {
			GsApp *addon = gs_app_list_index (addons, j);
			if (gs_app_get_to_be_installed (addon)) {
				const gchar *addon_source = gs_app_get_default_source (app);
				if (addon_source != NULL) {
					g_ptr_array_add (packages, g_strdup (addon_source));
					if (is_install)
						gs_app_set_state (addon, GS_APP_STATE_INSTALLING);
				}
			}
		}
	}

	rpm_proxy = gs_dnf5_rpm_rpm_proxy_new_sync (self->connection,
						    G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						    GS_DNF5_INTERFACE_RPM_DNF,
						    session_path,
						    cancellable,
						    error);
	success = rpm_proxy != NULL;
	if (success) {
		g_autoptr(GVariantBuilder) options_builder = NULL;

		g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (rpm_proxy), G_MAXINT);

		options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));

		/* NULL-terminate the array, to be like strv */
		g_ptr_array_add (packages, NULL);

		success = gs_dnf5_reset_transaction_sync (self, session_path, rpm_proxy, cancellable, error);

		if (success && is_install) {
			success = gs_dnf5_rpm_rpm_call_install_sync (rpm_proxy,
								     (const gchar * const *) packages->pdata,
								     g_variant_builder_end (options_builder),
								     cancellable,
								     error);
		} else if (success) {
			/* ignore package list for update, it's an offline action, which should update
			   everything possible, not only a specific package */
			const gchar * const all_packages[2] = { NULL, NULL };
			success = gs_dnf5_rpm_rpm_call_upgrade_sync (rpm_proxy,
								     all_packages,
								     g_variant_builder_end (options_builder),
								     cancellable,
								     error);
		}
		success = success && gs_dnf5_run_transaction (self, session_path, rpm_proxy, progress_app, progress_list,
							      is_install ? GS_DNF5_TRANSACTION_FLAG_NONE : GS_DNF5_TRANSACTION_FLAG_OFFLINE,
							      NULL, cancellable, error);
		if (success && !is_install) {
			/* ensure the update is finished with the 'reboot' action */
			g_autoptr(GsDnf5Offline) offline_proxy = NULL;

			offline_proxy = gs_dnf5_offline_proxy_new_sync (self->connection,
									G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
									GS_DNF5_INTERFACE_RPM_DNF,
									session_path,
									cancellable,
									error);
			success = offline_proxy != NULL;
			if (success) {
				g_autofree gchar *op_error_msg = NULL;
				gboolean op_success = FALSE;

				success = gs_dnf5_offline_call_set_finish_action_sync (offline_proxy, "reboot", &op_success, &op_error_msg, cancellable, error);
				if (success && !op_success) {
					success = FALSE;
					if (op_error_msg != NULL) {
						g_set_error_literal (error, GS_PLUGIN_ERROR, GS_PLUGIN_ERROR_FAILED, op_error_msg);
						g_prefix_error_literal (error, "Failed to set finish action: ");
					}
				} else if (!success) {
					gs_dnf5_convert_error (error);
					g_prefix_error_literal (error, "Failed to set finish action: ");
				}
			} else {
				gs_dnf5_convert_error (error);
				g_prefix_error_literal (error, "Failed to create Offline proxy: ");
			}
		}
	} else {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to create Rpm proxy: ");
	}

	gs_dnf5_close_session (self, session_manager, session_path);

	if (success) {
		gs_dnf5_mark_session_needs_reset (self);
	} else {
		gs_dnf5_convert_error (error);

		if (error != NULL && *error != NULL) {
			GsApp *app = NULL;
			if (gs_app_list_length (list) == 1)
				app = gs_app_list_index (list, 0);

			gs_dnf5_report_error (self, event_callback, event_user_data, app, *error, interactive);
		}
	}

	for (guint i = 0; i < gs_app_list_length (used_apps); i++) {
		GsApp *app = gs_app_list_index (used_apps, i);
		g_autoptr(GsAppList) addons = NULL;

		if (gs_app_get_state (app) == GS_APP_STATE_INSTALLING) {
			if (success) {
				gs_app_set_state (app, GS_APP_STATE_INSTALLED);
				gs_app_set_size_download (app, GS_SIZE_TYPE_VALID, 0);
			} else {
				gs_app_set_state_recover (app);
			}
		} else if (success) {
			/* apps prepared for offline update are downloaded */
			gs_app_set_size_download (app, GS_SIZE_TYPE_VALID, 0);
		}

		if (is_install)
			addons = gs_app_dup_addons (app);
		for (guint j = 0; addons != NULL && j < gs_app_list_length (addons); j++) {
			GsApp *addon = gs_app_list_index (addons, j);
			if (gs_app_get_state (addon) == GS_APP_STATE_INSTALLING) {
				if (success) {
					gs_app_set_state (addon, GS_APP_STATE_INSTALLED);
					gs_app_set_size_download (addon, GS_SIZE_TYPE_VALID, 0);
				} else {
					gs_app_set_state_recover (addon);
				}
			} else if (success) {
				/* apps prepared for offline update are downloaded */
				gs_app_set_size_download (app, GS_SIZE_TYPE_VALID, 0);
			}
		}
	}

	return success;
}

/* Run in @worker. */
static void
gs_dnf5_install_apps_thread_cb (GTask *task,
				gpointer source_object,
				gpointer task_data,
				GCancellable *cancellable)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	GsPlugin *plugin = GS_PLUGIN (self);
	GsPluginInstallAppsData *data = task_data;
	g_autoptr(GError) local_error = NULL;
	gboolean success;

	assert_in_worker (self);

	g_atomic_int_inc (&self->calling_rpm);

	success = gs_dnf5_install_update (plugin, data->apps, TRUE,
					  (data->flags & GS_PLUGIN_INSTALL_APPS_FLAGS_INTERACTIVE) != 0,
					  data->event_callback, data->event_user_data,
					  cancellable, &local_error);

	g_atomic_int_dec_and_test (&self->calling_rpm);

	if (success)
		g_task_return_boolean (task, TRUE);
	else
		g_task_return_error (task, g_steal_pointer (&local_error));
}

static void
gs_plugin_dnf5_install_apps_async (GsPlugin *plugin,
				   GsAppList *apps,
				   GsPluginInstallAppsFlags flags,
				   GsPluginProgressCallback progress_callback,
				   gpointer progress_user_data,
				   GsPluginEventCallback event_callback,
				   void *event_user_data,
				   GsPluginAppNeedsUserActionCallback app_needs_user_action_callback,
				   gpointer app_needs_user_action_data,
				   GCancellable *cancellable,
				   GAsyncReadyCallback callback,
				   gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	gboolean interactive = (flags & GS_PLUGIN_INSTALL_APPS_FLAGS_INTERACTIVE) != 0;
	g_autoptr(GTask) task = NULL;

	task = gs_plugin_install_apps_data_new_task (plugin, apps, flags, progress_callback, progress_user_data,
						     event_callback, event_user_data,
						     app_needs_user_action_callback, app_needs_user_action_data,
						     cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_install_apps_async);

	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_install_apps_thread_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_install_apps_finish (GsPlugin *plugin,
				    GAsyncResult *result,
				    GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

/* Run in @worker. */
static void
gs_dnf5_uninstall_apps_thread_cb (GTask *task,
				  gpointer source_object,
				  gpointer task_data,
				  GCancellable *cancellable)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	GsPlugin *plugin = GS_PLUGIN (self);
	GsPluginUninstallAppsData *data = task_data;
	g_autofree gchar *session_path = NULL;
	g_autoptr(GPtrArray) touched_apps = NULL;
	g_autoptr(GPtrArray) packages = NULL;
	g_autoptr(GsDnf5SessionManager) session_manager = NULL;
	g_autoptr(GsDnf5RpmRpm) rpm_proxy = NULL;
	g_autoptr(GsApp) progress_app = NULL;
	g_autoptr(GError) local_error = NULL;
	gboolean success;

	assert_in_worker (self);

	touched_apps = g_ptr_array_new_with_free_func (g_object_unref);
	packages = g_ptr_array_new_with_free_func (g_free);

	for (guint i = 0; i < gs_app_list_length (data->apps); i++) {
		g_autoptr(GsAppList) addons = NULL;
		GsApp *app = gs_app_list_index (data->apps, i);
		const gchar *source;
		if (!gs_app_has_management_plugin (app, plugin))
			continue;
		if (gs_app_get_kind (app) == AS_COMPONENT_KIND_REPOSITORY) {
			g_warn_if_reached ();
			continue;
		}
		source = gs_app_get_default_source (app);
		if (source == NULL) {
			g_task_return_new_error (task, GS_PLUGIN_ERROR, GS_PLUGIN_ERROR_NOT_SUPPORTED,
						 "remove not available");
			return;
		}
		g_ptr_array_add (packages, g_strdup (source));
		g_ptr_array_add (touched_apps, g_object_ref (app));
		gs_app_set_state (app, GS_APP_STATE_REMOVING);

		if (progress_app == NULL)
			progress_app = g_object_ref (app);

		addons = gs_app_dup_addons (app);
		for (guint j = 0; addons != NULL && j < gs_app_list_length (addons); j++) {
			GsApp *addon = gs_app_list_index (addons, j);
			if (gs_app_get_state (addon) == GS_APP_STATE_INSTALLED) {
				const gchar *addon_source = gs_app_get_default_source (addon);
				if (addon_source != NULL) {
					g_ptr_array_add (packages, g_strdup (addon_source));
					g_ptr_array_add (touched_apps, g_object_ref (addon));
					gs_app_set_state (addon, GS_APP_STATE_REMOVING);
				}
			}
		}
	}

	if (packages->len == 0) {
		g_task_return_boolean (task, TRUE);
		return;
	}

	session_path = gs_dnf5_open_session (self, GS_DNF5_RELEASEVER_DEFAULT, &session_manager, cancellable, &local_error);
	if (session_path == NULL) {
		gs_dnf5_convert_error (&local_error);
		g_task_return_error (task, g_steal_pointer (&local_error));
		return;
	}

	g_atomic_int_inc (&self->calling_rpm);

	rpm_proxy = gs_dnf5_rpm_rpm_proxy_new_sync (self->connection,
						    G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						    GS_DNF5_INTERFACE_RPM_DNF,
						    session_path,
						    cancellable,
						    &local_error);
	success = rpm_proxy != NULL;
	if (success) {
		g_autoptr(GVariantBuilder) options_builder = NULL;

		g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (rpm_proxy), G_MAXINT);

		options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
		g_variant_builder_add (options_builder, "{sv}", "strict",
				       g_variant_new_boolean (FALSE));

		/* NULL-terminate the array, to be like strv */
		g_ptr_array_add (packages, NULL);

		success = gs_dnf5_reset_transaction_sync (self, session_path, rpm_proxy, cancellable, &local_error);

		success = success &&
			  gs_dnf5_rpm_rpm_call_remove_sync (rpm_proxy,
							    (const gchar * const *) packages->pdata,
							    g_variant_builder_end (options_builder),
							    cancellable,
							    &local_error);
		success = success && gs_dnf5_run_transaction (self, session_path, rpm_proxy, progress_app, NULL, GS_DNF5_TRANSACTION_FLAG_NONE, NULL, cancellable, &local_error);
	} else {
		gs_dnf5_convert_error (&local_error);
		g_prefix_error_literal (&local_error, "Failed to create Rpm proxy: ");
	}

	gs_dnf5_close_session (self, session_manager, session_path);
	g_atomic_int_dec_and_test (&self->calling_rpm);

	if (success) {
		gs_dnf5_mark_session_needs_reset (self);
	} else {
		GsApp *app = NULL;
		if (gs_app_list_length (data->apps) == 1)
			app = gs_app_list_index (data->apps, 0);

		gs_dnf5_convert_error (&local_error);
		gs_dnf5_report_error (self, data->event_callback, data->event_user_data, app,
			local_error, (data->flags & GS_PLUGIN_UNINSTALL_APPS_FLAGS_INTERACTIVE) != 0);
	}

	for (guint i = 0; i < touched_apps->len; i++) {
		GsApp *app = g_ptr_array_index (touched_apps, i);
		if (gs_app_get_state (app) == GS_APP_STATE_REMOVING) {
			if (success)
				gs_app_set_state (app, GS_APP_STATE_UNKNOWN);
			else
				gs_app_set_state_recover (app);
		}
	}

	if (success)
		g_task_return_boolean (task, TRUE);
	else
		g_task_return_error (task, g_steal_pointer (&local_error));
}

static void
gs_plugin_dnf5_uninstall_apps_async (GsPlugin *plugin,
				     GsAppList *apps,
				     GsPluginUninstallAppsFlags flags,
				     GsPluginProgressCallback progress_callback,
				     gpointer progress_user_data,
				     GsPluginEventCallback event_callback,
				     void *event_user_data,
				     GsPluginAppNeedsUserActionCallback app_needs_user_action_callback,
				     gpointer app_needs_user_action_data,
				     GCancellable *cancellable,
				     GAsyncReadyCallback callback,
				     gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	gboolean interactive = (flags & GS_PLUGIN_UNINSTALL_APPS_FLAGS_INTERACTIVE) != 0;
	g_autoptr(GTask) task = NULL;

	task = gs_plugin_uninstall_apps_data_new_task (plugin, apps, flags, progress_callback, progress_user_data,
						       event_callback, event_user_data,
						       app_needs_user_action_callback, app_needs_user_action_data,
						       cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_uninstall_apps_async);

	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_uninstall_apps_thread_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_uninstall_apps_finish (GsPlugin *plugin,
				      GAsyncResult *result,
				      GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

/* Run in @worker. */
static void
gs_dnf5_update_apps_thread_cb (GTask *task,
			       gpointer source_object,
			       gpointer task_data,
			       GCancellable *cancellable)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	GsPlugin *plugin = GS_PLUGIN (self);
	GsPluginUpdateAppsData *data = task_data;
	g_autoptr(GError) local_error = NULL;
	gboolean success;

	assert_in_worker (self);

	g_atomic_int_inc (&self->calling_rpm);

	success = gs_dnf5_install_update (plugin, data->apps, FALSE,
					  (data->flags & GS_PLUGIN_UPDATE_APPS_FLAGS_INTERACTIVE) != 0,
					  data->event_callback, data->event_user_data,
					  cancellable, &local_error);

	g_atomic_int_dec_and_test (&self->calling_rpm);

	if (success)
		g_task_return_boolean (task, TRUE);
	else
		g_task_return_error (task, g_steal_pointer (&local_error));
}

static void
gs_plugin_dnf5_update_apps_async (GsPlugin *plugin,
				  GsAppList *apps,
				  GsPluginUpdateAppsFlags flags,
				  GsPluginProgressCallback progress_callback,
				  gpointer progress_user_data,
				  GsPluginEventCallback event_callback,
				  void *event_user_data,
				  GsPluginAppNeedsUserActionCallback app_needs_user_action_callback,
				  gpointer app_needs_user_action_data,
				  GCancellable *cancellable,
				  GAsyncReadyCallback callback,
				  gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	gboolean interactive = (flags & GS_PLUGIN_UPDATE_APPS_FLAGS_INTERACTIVE) != 0;
	g_autoptr(GTask) task = NULL;

	task = gs_plugin_update_apps_data_new_task (plugin, apps, flags, progress_callback, progress_user_data,
						    event_callback, event_user_data,
						    app_needs_user_action_callback, app_needs_user_action_data,
						    cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_update_apps_async);

	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_update_apps_thread_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_update_apps_finish (GsPlugin *plugin,
				   GAsyncResult *result,
				   GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

static gboolean
gs_dnf5_manage_repository_sync (GsPluginDnf5 *self,
				GsApp *repository,
				gboolean enable,
				GCancellable *cancellable,
				GError **error)
{
	g_autofree gchar *session_path = NULL;
	g_autoptr(GsDnf5SessionManager) session_manager = NULL;
	g_autoptr(GsDnf5RpmRepo) repo_proxy = NULL;
	gboolean success;

	if (!gs_app_has_management_plugin (repository, GS_PLUGIN (self)))
		return TRUE;

	if (gs_app_get_kind (repository) != AS_COMPONENT_KIND_REPOSITORY) {
		g_set_error (error,
			     GS_PLUGIN_ERROR,
			     GS_PLUGIN_ERROR_NOT_SUPPORTED,
			     "wrong app kind (%s) passed to manage_repository",
			     as_component_kind_to_string (gs_app_get_kind (repository)));
		return FALSE;
	}

	session_path = gs_dnf5_open_session (self, GS_DNF5_RELEASEVER_DEFAULT, &session_manager, cancellable, error);
	if (session_path == NULL) {
		gs_dnf5_convert_error (error);
		return FALSE;
	}

	repo_proxy = gs_dnf5_rpm_repo_proxy_new_sync (self->connection,
						      G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						      GS_DNF5_INTERFACE_RPM_DNF,
						      session_path,
						      cancellable,
						      error);
	success = repo_proxy != NULL;
	if (!success) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to create Rpm proxy: ");
	}

	if (success) {
		const gchar *ids[2] = { NULL, NULL };

		g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (repo_proxy), G_MAXINT);

		ids[0] = gs_app_get_id (repository);

		if (enable)
			success = gs_dnf5_rpm_repo_call_enable_sync (repo_proxy, ids, cancellable,  error);
		else
			success = gs_dnf5_rpm_repo_call_disable_sync (repo_proxy, ids, cancellable, error);

		if (success) {
			if (enable)
				gs_app_set_state (repository, GS_APP_STATE_INSTALLED);
			else
				gs_app_set_state (repository, GS_APP_STATE_AVAILABLE);
		} else {
			gs_dnf5_convert_error (error);
			g_prefix_error (error, "Failed to call %s: ", enable ? "enable" : "disable");
		}
	}

	gs_dnf5_close_session (self, session_manager, session_path);

	return success;
}

/* Run in @worker. */
static void
gs_dnf5_enable_repository_thread_cb (GTask *task,
				     gpointer source_object,
				     gpointer task_data,
				     GCancellable *cancellable)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	GsPluginManageRepositoryData *data = task_data;
	g_autoptr(GError) local_error = NULL;

	assert_in_worker (self);

	if (gs_dnf5_manage_repository_sync (self, data->repository, TRUE, cancellable, &local_error)) {
		g_task_return_boolean (task, TRUE);
	} else {
		gs_dnf5_convert_error (&local_error);
		g_task_return_error (task, g_steal_pointer (&local_error));
	}
}

static void
gs_plugin_dnf5_enable_repository_async (GsPlugin *plugin,
					GsApp *repository,
					GsPluginManageRepositoryFlags flags,
					GsPluginEventCallback event_callback,
					void *event_user_data,
					GCancellable *cancellable,
					GAsyncReadyCallback callback,
					gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	gboolean interactive = (flags & GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_INTERACTIVE) != 0;
	g_autoptr(GTask) task = NULL;

	task = gs_plugin_manage_repository_data_new_task (plugin, repository, flags, event_callback, event_user_data,
							  cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_enable_repository_async);

	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_enable_repository_thread_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_enable_repository_finish (GsPlugin *plugin,
					 GAsyncResult *result,
					 GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

/* Run in @worker. */
static void
gs_dnf5_disable_repository_thread_cb (GTask *task,
				      gpointer source_object,
				      gpointer task_data,
				      GCancellable *cancellable)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	GsPluginManageRepositoryData *data = task_data;
	g_autoptr(GError) local_error = NULL;

	assert_in_worker (self);

	if (gs_dnf5_manage_repository_sync (self, data->repository, FALSE, cancellable, &local_error)) {
		g_task_return_boolean (task, TRUE);
	} else {
		gs_dnf5_convert_error (&local_error);
		g_task_return_error (task, g_steal_pointer (&local_error));
	}
}

static void
gs_plugin_dnf5_disable_repository_async (GsPlugin *plugin,
					 GsApp *repository,
					 GsPluginManageRepositoryFlags flags,
					 GsPluginEventCallback event_callback,
					 void *event_user_data,
					 GCancellable *cancellable,
					 GAsyncReadyCallback callback,
					 gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	gboolean interactive = (flags & GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_INTERACTIVE) != 0;
	g_autoptr(GTask) task = NULL;

	task = gs_plugin_manage_repository_data_new_task (plugin, repository, flags, event_callback, event_user_data,
							  cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_disable_repository_async);

	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_disable_repository_thread_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_disable_repository_finish (GsPlugin *plugin,
					  GAsyncResult *result,
					  GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

static gboolean
gs_dnf5_pick_rpm_desktop_file_cb (GsPlugin *plugin,
				  GsApp *app,
				  const gchar *filename,
				  GKeyFile *key_file,
				  gpointer user_data)
{
	return strstr (filename, "/snapd/") == NULL &&
	       strstr (filename, "/snap/") == NULL &&
	       strstr (filename, "/flatpak/") == NULL &&
	       g_key_file_has_group (key_file, "Desktop Entry") &&
	       !g_key_file_has_key (key_file, "Desktop Entry", "X-Flatpak", NULL) &&
	       !g_key_file_has_key (key_file, "Desktop Entry", "X-SnapInstanceName", NULL);
}

static void
gs_plugin_dnf5_launch_async (GsPlugin *plugin,
			     GsApp *app,
			     GsPluginLaunchFlags flags,
			     GCancellable *cancellable,
			     GAsyncReadyCallback callback,
			     gpointer user_data)
{
	gs_plugin_app_launch_filtered_async (plugin, app, flags,
					     gs_dnf5_pick_rpm_desktop_file_cb, NULL,
					     cancellable,
					     callback, user_data);
}

static gboolean
gs_plugin_dnf5_launch_finish (GsPlugin *plugin,
			      GAsyncResult *result,
			      GError **error)
{
	return gs_plugin_app_launch_filtered_finish (plugin, result, error);
}

static gboolean
gs_dnf5_file_to_app_sync (GsPluginDnf5 *self,
			  GsAppList *list,
			  GFile *file,
			  GCancellable *cancellable,
			  GError **error)
{
	g_autofree gchar *content_type = NULL;
	g_autofree gchar *session_path = NULL;
	g_autoptr(GsDnf5SessionManager) session_manager = NULL;
	g_autoptr(GsDnf5RpmRpm) rpm_proxy = NULL;
	gboolean success = TRUE;
	const gchar *mimetypes[] = {
		"application/x-app-package",
		"application/x-deb",
		"application/vnd.debian.binary-package",
		"application/x-redhat-package-manager",
		"application/x-rpm",
		NULL };

	/* does this match any of the mimetypes we support */
	content_type = gs_utils_get_content_type (file, cancellable, error);
	if (content_type == NULL)
		return FALSE;
	if (!g_strv_contains (mimetypes, content_type))
		return TRUE;

	session_path = gs_dnf5_open_session (self, GS_DNF5_RELEASEVER_DEFAULT, &session_manager, cancellable, error);
	if (session_path == NULL) {
		gs_dnf5_convert_error (error);
		return FALSE;
	}

	rpm_proxy = gs_dnf5_rpm_rpm_proxy_new_sync (self->connection,
						    G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						    GS_DNF5_INTERFACE_RPM_DNF,
						    session_path,
						    cancellable,
						    error);
	success = rpm_proxy != NULL;
	if (!success) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to create Rpm proxy: ");
	}

	if (success) {
		g_autoptr(GVariantBuilder) options_builder = NULL;
		g_autoptr(GVariantBuilder) patterns_builder = NULL;
		g_autoptr(GVariant) result = NULL;
		g_autofree gchar *file_uri = NULL;

		file_uri = g_file_get_path (file);

		g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (rpm_proxy), G_MAXINT);

		patterns_builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
		g_variant_builder_add (patterns_builder, "s", file_uri);

		options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
		g_variant_builder_add (options_builder, "{sv}", "patterns",
				       g_variant_builder_end (patterns_builder));
		g_variant_builder_add (options_builder, "{sv}", "package_attrs",
				       gs_dnf5_dup_package_attrs ());
		g_variant_builder_add (options_builder, "{sv}", "with_provides",
				       g_variant_new_boolean (FALSE));
		g_variant_builder_add (options_builder, "{sv}", "with_filenames",
				       g_variant_new_boolean (FALSE));
		g_variant_builder_add (options_builder, "{sv}", "with_src",
				       g_variant_new_boolean (FALSE));
		g_variant_builder_add (options_builder, "{sv}", "scope",
				       g_variant_new_string ("all"));

		success = gs_dnf5_rpm_rpm_call_list_sync (rpm_proxy,
							  g_variant_builder_end (options_builder),
							  &result,
							  cancellable,
							  error);

		if (success) {
			g_autofree gchar *basename = g_file_get_basename (file);
			ReadPackageData rpd = { 0, };
			rpd.list = list;
			rpd.set_state = GS_APP_STATE_AVAILABLE_LOCAL;
			rpd.nevra_to_app = NULL;
			success = gs_dnf5_foreach_item (self, result, gs_dnf5_read_package_cb, &rpd, cancellable, error);

			if (success) {
				for (guint i = 0; i < gs_app_list_length (list); i++) {
					GsApp *app = gs_app_list_index (list, i);
					/* it's already downloaded, because it's a local file */
					gs_app_set_size_download (app, GS_SIZE_TYPE_VALID, 0);
					gs_app_set_bundle_kind (app, AS_BUNDLE_KIND_PACKAGE);
					gs_app_set_local_file (app, file);
					gs_app_set_metadata (app, "GnomeSoftware::packagename-value", basename);
				}
			}
		} else {
			gs_dnf5_convert_error (error);
			g_prefix_error_literal (error, "Failed to call list: ");
		}
	}

	gs_dnf5_close_session (self, session_manager, session_path);

	return success;
}

static void
gs_dnf5_file_to_app_thread_cb (GTask *task,
			       gpointer source_object,
			       gpointer task_data,
			       GCancellable *cancellable)
{
	g_autoptr(GsAppList) list = gs_app_list_new ();
	g_autoptr(GError) local_error = NULL;
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	GsPluginFileToAppData *data = task_data;

	if (gs_dnf5_file_to_app_sync (self, list, data->file, cancellable, &local_error)) {
		g_task_return_pointer (task, g_steal_pointer (&list), g_object_unref);
	} else if (local_error != NULL) {
		g_task_return_error (task, g_steal_pointer (&local_error));
	} else {
		g_task_return_pointer (task, g_steal_pointer (&list), g_object_unref);
	}
}

static void
gs_plugin_dnf5_file_to_app_async (GsPlugin *plugin,
				  GFile *file,
				  GsPluginFileToAppFlags flags,
				  GsPluginEventCallback event_callback,
				  void *event_user_data,
				  GCancellable *cancellable,
				  GAsyncReadyCallback callback,
				  gpointer user_data)
{
	g_autoptr(GTask) task = NULL;
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	gboolean interactive = (flags & GS_PLUGIN_FILE_TO_APP_FLAGS_INTERACTIVE) != 0;

	task = gs_plugin_file_to_app_data_new_task (plugin, file, flags, event_callback, event_user_data, cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_file_to_app_async);

	/* Queue a job to get the apps. */
	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_file_to_app_thread_cb, g_steal_pointer (&task));
}

static GsAppList *
gs_plugin_dnf5_file_to_app_finish (GsPlugin *plugin,
				   GAsyncResult *result,
				   GError **error)
{
	return g_task_propagate_pointer (G_TASK (result), error);
}

static gboolean
gs_dnf5_system_upgrade_sync (GsPluginDnf5 *self,
			     const gchar *releasever,
			     GsApp *progress_app,
			     gboolean download_only,
			     GCancellable *cancellable,
			     GError **error)
{
	g_autofree gchar *session_path = NULL;
	g_autoptr(GsAppList) list = NULL;
	g_autoptr(GsDnf5SessionManager) session_manager = NULL;
	g_autoptr(GsDnf5Offline) offline_proxy = NULL;
	g_autoptr(GsDnf5RpmRpm) rpm_proxy = NULL;
	gboolean success;

	session_path = gs_dnf5_open_session (self, releasever, &session_manager, cancellable, error);
	if (session_path == NULL) {
		gs_dnf5_convert_error (error);
		return FALSE;
	}

	rpm_proxy = gs_dnf5_rpm_rpm_proxy_new_sync (self->connection,
						    G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
						    GS_DNF5_INTERFACE_RPM_DNF,
						    session_path,
						    cancellable,
						    error);
	success = rpm_proxy != NULL;
	if (!success) {
		gs_dnf5_convert_error (error);
		g_prefix_error_literal (error, "Failed to create Rpm proxy: ");
	}

	if (success) {
		offline_proxy = gs_dnf5_offline_proxy_new_sync (self->connection,
								G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
								GS_DNF5_INTERFACE_RPM_DNF,
								session_path,
								cancellable,
								error);
		success = offline_proxy != NULL;
		if (!success) {
			gs_dnf5_convert_error (error);
			g_prefix_error_literal (error, "Failed to create Offline proxy: ");
		}
	}

	if (success) {
		gboolean pending = FALSE;
		g_autoptr(GVariant) transaction_status = NULL;

		if (gs_dnf5_offline_call_get_status_sync (offline_proxy, &pending, &transaction_status, cancellable, error) &&
		    pending && transaction_status != NULL) {
			const gchar *target_releasever = NULL;
			gboolean found = g_variant_lookup (transaction_status, "target_releasever", "&s", &target_releasever);
			if (found && g_strcmp0 (target_releasever, releasever) == 0) {
				g_debug ("There is prepared an upgrade for version '%s' already, skipping preparation", releasever);
				gs_dnf5_close_session (self, session_manager, session_path);
				return TRUE;
			} else if (found && target_releasever != NULL) {
				g_debug ("There is prepared an update for version '%s', but needs version '%s', redo preparation", target_releasever, releasever);
			}
		}
	}

	if (success) {
		g_autoptr(GVariantBuilder) options_builder = NULL;
		g_autoptr(GVariant) result = NULL;

		g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (rpm_proxy), G_MAXINT);

		options_builder = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));

		success = gs_dnf5_reset_transaction_sync (self, session_path, rpm_proxy, cancellable, error);

		success = success &&
			  gs_dnf5_rpm_rpm_call_system_upgrade_sync (rpm_proxy,
								    g_variant_builder_end (options_builder),
								    cancellable,
								    error);

		if (success) {
			success = gs_dnf5_run_transaction (self, session_path, rpm_proxy, progress_app, NULL, GS_DNF5_TRANSACTION_FLAG_OFFLINE, NULL, cancellable, error);

			if (success && !download_only) {
				g_autofree gchar *op_error_msg = NULL;
				gboolean op_success = FALSE;

				/* ensure the upgrade is finished with the 'reboot' action */
				success = gs_dnf5_offline_call_set_finish_action_sync (offline_proxy, "reboot", &op_success, &op_error_msg, cancellable, error);
				if (success && !op_success) {
					success = FALSE;
					if (op_error_msg != NULL) {
						g_set_error_literal (error, GS_PLUGIN_ERROR, GS_PLUGIN_ERROR_FAILED, op_error_msg);
						g_prefix_error_literal (error, "Failed to set finish action: ");
					}
				} else if (!success) {
					gs_dnf5_convert_error (error);
					g_prefix_error_literal (error, "Failed to set finish action: ");
				}
			}
		} else {
			gs_dnf5_convert_error (error);
			g_prefix_error_literal (error, "Failed to call system_upgrade: ");
		}
	}

	gs_dnf5_close_session (self, session_manager, session_path);

	return success;
}

/* Run in @worker. */
static void
gs_dnf5_download_upgrade_thread_cb (GTask *task,
				    gpointer source_object,
				    gpointer task_data,
				    GCancellable *cancellable)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	GsPluginDownloadUpgradeData *data = task_data;
	g_autoptr(GError) local_error = NULL;

	assert_in_worker (self);

	gs_app_set_state (data->app, GS_APP_STATE_DOWNLOADING);

	if (gs_dnf5_system_upgrade_sync (self, gs_app_get_version (data->app), data->app, TRUE, cancellable, &local_error)) {
		gs_app_set_state (data->app, GS_APP_STATE_UPDATABLE);
		g_task_return_boolean (task, TRUE);
	} else {
		gs_app_set_state (data->app, GS_APP_STATE_AVAILABLE);
		gs_dnf5_convert_error (&local_error);
		gs_dnf5_report_error (self, data->event_callback, data->event_user_data, data->app,
			local_error, (data->flags & GS_PLUGIN_DOWNLOAD_UPGRADE_FLAGS_INTERACTIVE) != 0);
		g_task_return_error (task, g_steal_pointer (&local_error));
	}
}

static void
gs_plugin_dnf5_download_upgrade_async (GsPlugin *plugin,
				       GsApp *app,
				       GsPluginDownloadUpgradeFlags flags,
				       GsPluginEventCallback event_callback,
				       void *event_user_data,
				       GCancellable *cancellable,
				       GAsyncReadyCallback callback,
				       gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	g_autoptr(GTask) task = NULL;
	gboolean interactive = (flags & GS_PLUGIN_DOWNLOAD_UPGRADE_FLAGS_INTERACTIVE) != 0;

	task = gs_plugin_download_upgrade_data_new_task (plugin, app, flags, event_callback, event_user_data, cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_download_upgrade_async);

	/* only process this app if was created by this plugin */
	if (!gs_app_has_management_plugin (app, plugin)) {
		g_task_return_boolean (task, TRUE);
		return;
	}

	/* check is distro-upgrade */
	if (gs_app_get_kind (app) != AS_COMPONENT_KIND_OPERATING_SYSTEM) {
		g_task_return_boolean (task, TRUE);
		return;
	}

	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_download_upgrade_thread_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_download_upgrade_finish (GsPlugin *plugin,
					GAsyncResult *result,
					GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

/* Run in @worker. */
static void
gs_dnf5_trigger_upgrade_thread_cb (GTask *task,
				   gpointer source_object,
				   gpointer task_data,
				   GCancellable *cancellable)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (source_object);
	GsPluginTriggerUpgradeData *data = task_data;
	g_autoptr(GError) local_error = NULL;

	assert_in_worker (self);

	gs_app_set_state (data->app, GS_APP_STATE_PENDING_INSTALL);

	if (gs_dnf5_system_upgrade_sync (self, gs_app_get_version (data->app), data->app, FALSE, cancellable, &local_error)) {
		gs_app_set_state (data->app, GS_APP_STATE_UPDATABLE);
		g_task_return_boolean (task, TRUE);
	} else {
		gs_app_set_state (data->app, GS_APP_STATE_AVAILABLE);
		gs_dnf5_convert_error (&local_error);
		g_task_return_error (task, g_steal_pointer (&local_error));
	}
}

static void
gs_plugin_dnf5_trigger_upgrade_async (GsPlugin *plugin,
				      GsApp *app,
				      GsPluginTriggerUpgradeFlags flags,
				      GCancellable *cancellable,
				      GAsyncReadyCallback callback,
				      gpointer user_data)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (plugin);
	g_autoptr(GTask) task = NULL;
	gboolean interactive = (flags & GS_PLUGIN_TRIGGER_UPGRADE_FLAGS_INTERACTIVE) != 0;

	task = gs_plugin_trigger_upgrade_data_new_task (plugin, app, flags, cancellable, callback, user_data);
	g_task_set_source_tag (task, gs_plugin_dnf5_trigger_upgrade_async);

	/* only process this app if was created by this plugin */
	if (!gs_app_has_management_plugin (app, plugin)) {
		g_task_return_boolean (task, TRUE);
		return;
	}

	/* check is distro-upgrade */
	if (gs_app_get_kind (app) != AS_COMPONENT_KIND_OPERATING_SYSTEM) {
		g_task_return_boolean (task, TRUE);
		return;
	}

	gs_app_set_state (app, GS_APP_STATE_PENDING_INSTALL);

	gs_worker_thread_queue (self->worker, gs_dnf5_get_priority_for_interactivity (interactive),
				gs_dnf5_trigger_upgrade_thread_cb, g_steal_pointer (&task));
}

static gboolean
gs_plugin_dnf5_trigger_upgrade_finish (GsPlugin *plugin,
				       GAsyncResult *result,
				       GError **error)
{
	return g_task_propagate_boolean (G_TASK (result), error);
}

static void
gs_plugin_dnf5_dispose (GObject *object)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (object);

	if (self->rpm_transaction_watch_id != 0) {
		g_bus_unwatch_name (self->rpm_transaction_watch_id);
		self->rpm_transaction_watch_id = 0;
	}

	g_clear_object (&self->worker);
	g_clear_object (&self->connection);
	g_clear_object (&self->rpm_transaction_proxy);

	g_mutex_lock (&self->session_data.mutex);
	if (self->session_data.autoclose_timer) {
		g_source_remove (self->session_data.autoclose_timer);
		self->session_data.autoclose_timer = 0;
	}

	if (self->session_data.proxy != NULL) {
		gs_dnf5_close_session_real (self->session_data.proxy, self->session_data.object_path);

		g_clear_object (&self->session_data.proxy);
		g_clear_pointer (&self->session_data.object_path, g_free);
	}
	g_mutex_unlock (&self->session_data.mutex);

	g_mutex_lock (&self->dependency_sizes_data.mutex);
	g_clear_pointer (&self->dependency_sizes_data.apps, g_ptr_array_unref);
	g_clear_object (&self->dependency_sizes_data.cancellable);
	g_mutex_unlock (&self->dependency_sizes_data.mutex);

	G_OBJECT_CLASS (gs_plugin_dnf5_parent_class)->dispose (object);
}

static void
gs_plugin_dnf5_finalize (GObject *object)
{
	GsPluginDnf5 *self = GS_PLUGIN_DNF5 (object);

	g_mutex_clear (&self->session_data.mutex);
	g_cond_clear (&self->session_data.cond);
	g_mutex_clear (&self->dependency_sizes_data.mutex);

	G_OBJECT_CLASS (gs_plugin_dnf5_parent_class)->finalize (object);
}

static void
gs_plugin_dnf5_class_init (GsPluginDnf5Class *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GsPluginClass *plugin_class = GS_PLUGIN_CLASS (klass);

	object_class->dispose = gs_plugin_dnf5_dispose;
	object_class->finalize = gs_plugin_dnf5_finalize;

	plugin_class->setup_async = gs_plugin_dnf5_setup_async;
	plugin_class->setup_finish = gs_plugin_dnf5_setup_finish;
	plugin_class->shutdown_async = gs_plugin_dnf5_shutdown_async;
	plugin_class->shutdown_finish = gs_plugin_dnf5_shutdown_finish;
	plugin_class->list_apps_async = gs_plugin_dnf5_list_apps_async;
	plugin_class->list_apps_finish = gs_plugin_dnf5_list_apps_finish;
	plugin_class->refine_async = gs_plugin_dnf5_refine_async;
	plugin_class->refine_finish = gs_plugin_dnf5_refine_finish;
	plugin_class->refresh_metadata_async = gs_plugin_dnf5_refresh_metadata_async;
	plugin_class->refresh_metadata_finish = gs_plugin_dnf5_refresh_metadata_finish;
	plugin_class->update_apps_async = gs_plugin_dnf5_update_apps_async;
	plugin_class->update_apps_finish = gs_plugin_dnf5_update_apps_finish;
	plugin_class->enable_repository_async = gs_plugin_dnf5_enable_repository_async;
	plugin_class->enable_repository_finish = gs_plugin_dnf5_enable_repository_finish;
	plugin_class->disable_repository_async = gs_plugin_dnf5_disable_repository_async;
	plugin_class->disable_repository_finish = gs_plugin_dnf5_disable_repository_finish;
	plugin_class->install_apps_async = gs_plugin_dnf5_install_apps_async;
	plugin_class->install_apps_finish = gs_plugin_dnf5_install_apps_finish;
	plugin_class->uninstall_apps_async = gs_plugin_dnf5_uninstall_apps_async;
	plugin_class->uninstall_apps_finish = gs_plugin_dnf5_uninstall_apps_finish;
	plugin_class->launch_async = gs_plugin_dnf5_launch_async;
	plugin_class->launch_finish = gs_plugin_dnf5_launch_finish;
	plugin_class->file_to_app_async = gs_plugin_dnf5_file_to_app_async;
	plugin_class->file_to_app_finish = gs_plugin_dnf5_file_to_app_finish;
	plugin_class->download_upgrade_async = gs_plugin_dnf5_download_upgrade_async;
	plugin_class->download_upgrade_finish = gs_plugin_dnf5_download_upgrade_finish;
	plugin_class->trigger_upgrade_async = gs_plugin_dnf5_trigger_upgrade_async;
	plugin_class->trigger_upgrade_finish = gs_plugin_dnf5_trigger_upgrade_finish;
}

static void
gs_plugin_dnf5_init (GsPluginDnf5 *self)
{
	GsPlugin *plugin = GS_PLUGIN (self);

	g_mutex_init (&self->session_data.mutex);
	g_cond_init (&self->session_data.cond);
	g_mutex_init (&self->dependency_sizes_data.mutex);

	/* getting app properties from appstream is quicker */
	gs_plugin_add_rule (plugin, GS_PLUGIN_RULE_RUN_AFTER, "appstream");

	/* like appstream, we need the icon plugin to load cached icons into pixbufs */
	gs_plugin_add_rule (plugin, GS_PLUGIN_RULE_RUN_BEFORE, "icons");

	/* prioritize over packagekit */
	gs_plugin_add_rule (plugin, GS_PLUGIN_RULE_CONFLICTS, "packagekit");
}

GType
gs_plugin_query_type (void)
{
	return GS_TYPE_PLUGIN_DNF5;
}
